"""Provides an engine for driving measurements. The design pattern is to
be as modular as possible. The engine drives a set of measurements and
the measurements themselves as well as the filters, etc are all defined
by the user via the experiment and the factory. The engine should not
include any filtering or measuring within its code base.

The goal of this engine is to be as flexible as possible so no other
engine is needed. However, the framework architecutal approach remains
consistent in that this too can be swapped out by creation of another
type in the factory and use a scenario to invoke.
"""


import logging
from typing import Union, List, Any
from adgtk.common import (
    ComponentDef,
    FactoryBlueprint,
    ArgumentSetting,
    FolderManager,
    ArgumentType)
from adgtk.tracking import MetricTracker
from adgtk.factory.component import ObjectFactory
from adgtk.journals import ExperimentJournal
from .base import SupportsMeasSetOps
from adgtk.tracking import MetricTracker

# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/instrumentation/test_engine.py


# ----------------------------------------------------------------------
# Module Constants
# ----------------------------------------------------------------------
DEBUG_TO_CONSOLE = False
LOW_DATA_THRESHOLD: Union[int, None] = None  # Set to None to disable


# ----------------------------------------------------------------------
# Engine
# ----------------------------------------------------------------------


class MeasurementEngine:
    """Responsible for providing a wrapper around a group of measurement
    sets. It also exposes a measurement tracker for consumption of the
    measurement results. The overall design is lightweight, it invokes
    the measurement sets for interacting with their measurements."""

    description = "A measurement Engine is a collection of measurement sets."
    blueprint: FactoryBlueprint = {
        "group_label": "measurement-engine",
        "type_label": "simple",
        "introduction": """The Measurement Engine is a collection of measurement sets. The engine is responsible for invoking the measurements and storing the results. The engine is also responsible for exporting the results to a report. The engine is a lightweight wrapper around the measurement sets. You will need to configure the following:

        - The name of this engine
        - a measurement set (which is a collection of measurements)

note: The measurement Engine can have one or more measurement sets.""",
        "arguments": {
            "name": ArgumentSetting(
                help_str="the name of the measurement engine",
                default_value="main-engine",
                argument_type=ArgumentType.STRING),            
            "measurement_set_def": ArgumentSetting(
                help_str="A measurement set for this engine",
                argument_type=ArgumentType.LIST,
                group_label="measurement-set",
                list_intro="In order to create a measurement set, you will need to define the measurements. The measurements are the individual components that will be used to measure the data. The measurements are defined in the measurement set. The measurement set is a collection of measurements. You can have one or more measurement sets. This section will allow you to define the measurements for the measurement set.",
                list_min=1,
                list_arg_type=ArgumentType.BLUEPRINT,
                list_group_label="measurement-set")
        }
    }

    def __init__(
        self,
        factory: ObjectFactory,
        measurement_set_def: List[ComponentDef] | None = None,
        name: str = "default",
        journal: ExperimentJournal | None = None
    ):
        """Initializes a new instance of the Measurement Engine.

        :param factory: The component factory
        :type factory: ObjectFactory
        :param measurement_set_def: measurement set(s), defaults to None
        :type measurement_set_def: List[ComponentDef] | None, optional
        :param name: Name of the engine, defaults to "default"
        :type name: str, optional
        :param journal: The experiment journal, defaults to None
        :type journal: ExperimentJournal | None, optional
        """
        self.metric_tracker: Union[MetricTracker, None] = None
        self.name = name
        self._journal = journal
        self.factory = factory
        self._meas_sets: dict[str, SupportsMeasSetOps] = {}
        if measurement_set_def is None:
            measurement_set_def = []

        for meas_set_def in measurement_set_def:
            self.add_measurement_set(meas_set_def)

        if journal is not None:
            # tell the journal that the engine exists
            journal.register_meas_engine(self)
        else:
            logging.info("Measurement Engine created without a journal")

    def get_measurement_set_names(self) -> list:
        """Returns the names of the registered measurement sets. Usage
        includes for generating reports and needing to iterate the
        measurement tracker.

        :return: A list of the measurement set names
        :rtype: list
        """
        return list(self._meas_sets.keys())

    def add_measurement_set(self, meas_set_def: ComponentDef) -> None:
        """adds a new measurement set into the Engine. Used on init and
        if needed during experiment execution can be invoked to expand
        the data.

        :param meas_set_def: the measurement set to create and add
        :type meas_set_def: ComponentDef
        """
        meas_set: SupportsMeasSetOps
        meas_set = self.factory.create(meas_set_def)

        # update measurement_set engine name (reporting, exports, etc)
        meas_set.engine_name = self.name

        if meas_set.name in self._meas_sets:
            msg = "Attempted to insert a duplicate MeasurementSet "
            msg += f"{meas_set.name} into MeasurementEngine {self.name}"
            msg += ". Request ignored. No changes made."
            return None

        self._meas_sets[meas_set.name] = meas_set

    def remove_measurement_set(
        self,
        name: str,
        remove_data: bool = False
    ) -> None:
        """Removes a measurement set if one exists along

        :param name: The name of the measurement set to remove
        :type name: str
        :param remove_data: Also remove the tracking data,
            defaults to False
        :type remove_data: bool, optional
        """

        if name in self._meas_sets:
            _ = self._meas_sets.pop(name)

        if remove_data:
            raise NotImplementedError("Code more here!")

    def update_stopwords(self, stopwords: List[str]) -> None:
        """Iterates through all the different sets and updates stopwords
        as needed.

        :param stopwords: The stopwords used by measurements.
        :type stopwords: List[str]
        """
        for meas_set in self._meas_sets.values():
            meas_set.update_stopwords(stopwords)
            msg = f"Updated stopwords for {self.name}.{meas_set.name}"
            logging.info(msg)

    def update_metric_tracker(self, metric_tracker: MetricTracker):
        """Update the metric tracker"""
        self.metric_tracker = metric_tracker
        for meas_set in self._meas_sets.values():
            meas_set.register_metric_tracker(metric_tracker)

        # TODO: update journal(s)

    def perform_measurements(self, data: Any) -> None:
        """Executes the measurements currently active. All results are
        store in the measurement tracker.

        :param data: the data to measure
        :type data: Any
        """
        # ensure we have a metric tracker if one does not already exist
        if self.metric_tracker is None:
            metric_tracker = MetricTracker()
            self.update_metric_tracker(metric_tracker)

        meas_set: SupportsMeasSetOps
        for meas_set in self._meas_sets.values():
            meas_set.perform_measurements(data=data)


    def create_html_and_export(
        self,
        experiment_name: str,
        settings_file_override: Union[str, None] = None,
        header: int = 2,
        base_url: str = "http://127.0.0.1:8000"
    ) -> str:
        """Creates local files and returns HTML that can be used by the
        reports. In addition, exports data, builds images, etc in order
        to save and report.

        
        :param experiment_name: The name of the experiment
        :type experiment_name: str
        :param settings_file_override: The settings file to use, defaults to None
        :type settings_file_override: Union[str, None], optional
        :param header: The header level to use, defaults to 2
        :type header: int, optional
        :param base_url: The base URL to use, defaults to "http://127.0.0.01:8000
        :type base_url: str, optional
        :return: The HTML to use in the report
        :rtype: str
        
        """
    
                
        html = f"""
        <div>
            <h{header}>Measurement Engine {self.name} </h{header}>
        """
        folders = FolderManager(
            name=experiment_name, settings_file_override=settings_file_override)

        meas_set: SupportsMeasSetOps
        for meas_set in self._meas_sets.values():
            # confirm/double check engine and override as needed
            meas_set.engine_name = self.name
            # now generate HTML
            html += meas_set.create_html_and_export(
                header=3, folders=folders, base_url=base_url)
        html += "</div>"
        return html



"""Provides a common place for default values. Helps to avoid circular
importing."""
from typing import Literal

# Experiment Folders
EXP_DEF_DIR = "blueprints"
BATCH_DEF_DIR = "batch"
SHARED_MODEL_DIR = "models"

# Logging
LOG_DIR = "logs"
SCENARIO_LOGGER_NAME = "SCENARIO"


"""Common and useful outside of the module."""
from .structure import (
    BlueprintQuestion,
    FactoryOrder,    
    SupportedCaptureTypes,
    SupportedChoiceTypes,
    SupportsFactory
)
from .component import register_to_factory


# mypy: ignore-errors
"""Agent base provides the foundational components for the module.
"""
# pylint: disable=unused-argument

from typing import Union, Protocol, runtime_checkable
from adgtk.common import FactoryBlueprint
from adgtk.components.base import State, Action
from adgtk.components.environment import Environment


@runtime_checkable
class Agent(Protocol):
    """The Agent Protocol"""
    blueprint: FactoryBlueprint

    def engage(self, state: State) -> Action:
        """Engages the Policy for a single state. The goal is to exploit
        the policy not explore/train. This method will set the policy
        into exploit mode versus training mode.

        :param state: The state to invoke the Policy with
        :type state: State
        :return: The action based on the Policy
        :rtype: Action
        """

    def train(
        self,
        experiment_name: str,
        train_environment: Environment,
        val_environment: Union[Environment, None] = None,
        test_environment: Union[Environment, None] = None,
        epochs: int = 1,
    ) -> None:
        """Explores an environment and trains the provided policy to
        learn how to predict match versus non-match for the entities.

        :param experiment_name: The name of the experiment.
        :type experiment_name: str
        :param train_environment: The env for training
        :type train_environment: Environment
        :param val_environment: validation env
        :type val_environment: Environment, optional
        :param test_environment: The test env
        :type test_environment: Environment, optional
        :param epochs:  epochs to train, defaults to 1
        :type epochs: int, optional
        """



"""The Presentation module is focused on defining Protocols for 
presenting data. These protocol(s) are intented to be used by the
components responsible for presenting data in a specific format.
"""


from typing import Protocol, runtime_checkable

# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/data/test_presentation.py


@runtime_checkable
class PresentationFormat(Protocol):
    """Provides a base for a presentation """

    def present(self, data: dict) -> str:
        """Presents data based on its configuration

        :param data: the data to be presented
        :type data: dict
        :return: a string in the format configured
        :rtype: str
        """



"""The Manager module is responsible for managing the lifecycle of a
scenario.
"""

import importlib
import os
import sys
import logging
from typing import Literal, Union, Any
import toml
import yaml
from adgtk.common import (
    FactoryBlueprint,
    ComponentDef,
    DuplicateFactoryRegistration,
    InvalidScenarioState,
    convert_exp_def_to_string,
    FolderManager)
from adgtk.factory import ObjectFactory
from adgtk.journals import ExperimentJournal
from adgtk.instrumentation import measurement_register_list
from .base import Scenario, SCENARIO_GROUP_LABEL

# ----------------------------------------------------------------------
# py -m pytest test/scenario/test_scenario_manager.py

# ----------------------------------------------------------------------
# Module configs
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
# Management : running experiments and creating blueprints
# ----------------------------------------------------------------------

class ScenarioManager:
    """Manages all aspects of a scenario."""

    def __init__(
        self,
        experiment_definition_dir: str = "experiment-definition",
        settings_file_override: Union[str, None] = None,
        load_base: bool = True,
        load_user_modules: Union[list, None] = None,
    ) -> None:
        if load_user_modules is None:
            load_user_modules = []

        self.experiment_definition_dir = experiment_definition_dir
        self.active_scenario: Union[Scenario, None] = None
        self.settings_file_override = settings_file_override
        self.load_user_modules_errs: list[str] = []
        # setup internal objects
        self._journal = ExperimentJournal()
        self._factory = ObjectFactory(
            journal=self._journal,
            settings_file_override=settings_file_override)
        self.active_scenario_name = "SCENARIO-MGR-NOT-SET"

        # Load the core such as Measurement Engine
        self._load_core_components()

        for user_module in load_user_modules:
            self._load_user_components(user_module)

    def __str__(self):
        report = "Scenario Manager\n"
        report += "================\n"
        report += f"{self._factory}"
        return report

    def group_report(self, group_label: str) -> str:
        """Creates a report string for a group. Primary use is in the
        command line interface.

        :param group_label: The group to filter on
        :type group_label: str
        :return: a report showing the group members
        :rtype: str
        """
        return self._factory.group_report(group_label)

    def _load_experiment_from_file(self, file_w_path: str) -> ComponentDef:
        """Loads an experiment definition from disk

        Args:
            file_w_path (str): The file containing the definition

        Raises:
            NotImplementedError: More work needed here
            FileNotFoundError: Unable to load the file

        Returns:
            ComponentDef: The settings for the experiment
        """

        # current_directory = os.getcwd()
        # print(f"Current Directory: {current_directory}")

        if os.path.exists(os.path.join(file_w_path)):
            with open(file_w_path, "r", encoding="utf-8") as infile:
                if file_w_path.lower().endswith(".toml"):
                    exp_def = toml.load(infile)
                elif file_w_path.lower().endswith(".yaml"):
                    exp_def = yaml.safe_load(infile)
                else:
                    msg = f"Unexpected file format for {file_w_path}"
                    logging.error(msg)
                    raise InvalidScenarioState(msg)
                if "comments" in exp_def:
                    start_msg = f"Starting: {file_w_path}"
                    logging.info(start_msg)
                    logging.info(exp_def["comments"])
                    self._journal.add_entry(
                        entry_type="comment",
                        component="scenario",
                        entry_text=exp_def["comments"])
                if "configuration" in exp_def:
                    return exp_def["configuration"]

        elif os.path.exists(file_w_path):
            raise NotImplementedError("DEV needed")
        else:
            raise FileNotFoundError(
                f"Unable to find experiment definition: {file_w_path}")

        # should not happen.
        return ComponentDef(
            type_label="error",
            group_label="error",
            arguments={})

    def _load_user_components(self, user_module: str) -> int:
        """Loads a user module and the registers the components from the
        register_list.

        :param user_module: The module to load
        :type user_module: str
        :raises ModuleNotFoundError: user module not found
        :raises AttributeError: register_list not found
        :return: The number of items registered
        :rtype: int
        """
        count = 0

        # ensures the cwd is loaded to the path so that the module the
        # user wants to load is found. W/out this you get a module not
        # found error.
        if sys.path[0] != '':
            sys.path.insert(0, '')
            # uncomment to troubleshoot
            # print(sys.path)

        # moving to an improved design where we skip this step so as to
        # fail with an error instead of gracefully skipping.

        # user_spec = importlib.util.find_spec(user_module)
        # if user_spec is None:
        #     msg = f"Module {user_module} not found at {os.getcwd()}"
        #     print(msg)
        #     logging.error(msg)
        #     return 0

        # attempt to load the module
        try:
            # loaded_module = importlib.import_module(user_spec.name)
            loaded_module = importlib.import_module(user_module)
            msg = f"ScenarioManager Loaded module {user_module}"
            logging.info(msg)
        except ModuleNotFoundError as e:
            # moved to not using user_spec so that the user can get a
            # better error message than a default.
            print("USER MODULE ERRORS")
            print("==================")
            msg = f"ERROR: User Module [{user_module}]"
            print(msg)
            raise e

            # TODO: cleanup once sure this is not needed
            # should not happen as we found the spec
            msg = f"ERROR: User Module [{user_module}] not found"
            self.load_user_modules_errs.append(msg)
            logging.error(msg)
            print("")
            sys.exit(1)
        try:
            user_register_list = loaded_module.register_list
        except AttributeError:
            msg = f"unable to find register_list in {user_module}"
            logging.error(msg)
            print("")
            self.load_user_modules_errs.append(msg)
            return 0

        for creator in user_register_list:
            self.register(creator=creator)
            count += 1

        return count

    def _load_factories(
        self,
        creator: Any,
        group_label_override: Union[str, None] = None,
        type_label_override: Union[str, None] = None
    ) -> None:

        # Load component
        self._factory.register(
            group_label_override=group_label_override,
            type_label_override=type_label_override,
            creator=creator)

    def _load_core_components(self):
        """Loads components that should be registered regardless. For
        example components that perform measurements."""

        # for measurements
        for creator in measurement_register_list:
            self.register(creator=creator)

    def load_experiment(self, experiment_name: str) -> None:
        """Loads and experiment

        :param experiment_name: The name of the file
        :type experiment_name: str
        """
        file_w_path = os.path.join(
            self.experiment_definition_dir, experiment_name)
        alt1 = f"{file_w_path}.toml"
        alt2 = f"{file_w_path}.yaml"
        exp_def: Union[ComponentDef, None] = None
        if os.path.exists(file_w_path):
            msg = f"Loaded {experiment_name}"
            logging.info(msg)
            exp_def = self._load_experiment_from_file(file_w_path)
        elif os.path.exists(alt1):
            msg = f"expanded from {experiment_name} to {alt1} and loaded"
            logging.info(msg)
            exp_def = self._load_experiment_from_file(alt1)
        elif os.path.exists(alt2):
            msg = f"expanded from {experiment_name} to {alt2} and loaded"
            logging.info(msg)
            exp_def = self._load_experiment_from_file(alt2)
        else:
            msg = f"Unable to find experiment: {experiment_name}. Check file."
            print(msg)
            logging.fatal(msg)
            sys.exit(1)            

        # folders should not have the extension so it is stripped
        # It should also be lower case so as to keep things predictable
        experiment_name = experiment_name.lower()
        if experiment_name.endswith(".toml") or \
                experiment_name.endswith(".yaml"):
            experiment_name = experiment_name[:-5]
        self.active_scenario_name = experiment_name

        # update folder manager so that objects created know where to
        # write their files/updates/etc to.
        self._factory.update_folder_manager(name=experiment_name)
        # And update the Journal(s)
        exp_tree = convert_exp_def_to_string(exp_def=exp_def)
        self._journal.add_entry(
            entry_text=exp_tree,
            entry_type="scenario_def",
            include_timestamp=False)

        # and then create the scenario
        self.active_scenario = self._factory.create(exp_def)

    def preview_experiment(self, to_console: bool = True) -> None:
        """Initiates the scenario preview.
        :param to_console: Print the tree to the console
        :type to_console: bool
        :raises InvalidScenarioState: no scenario loaded
        """
        folder_mgr = FolderManager(
            name=self.active_scenario_name,
            clear_and_rebuild=False)

        if self.active_scenario is not None:
            self._journal.generate_preview(
                experiment_name=self.active_scenario_name,
                experiment_folder=folder_mgr.base_folder)

        else:
            msg = "No loaded scenario to preview"
            logging.error(msg)
            raise InvalidScenarioState(msg)

    def run_experiment(self) -> None:
        """Initiates the scenario run

        :raises InvalidScenarioState: no scenario loaded
        """
        if self.active_scenario is not None:
            self.active_scenario.execute(self.active_scenario_name)
        else:
            msg = "No loaded scenario to run"
            logging.error(msg)
            raise InvalidScenarioState(msg)

        # And report
        folder_mgr = FolderManager(
            name=self.active_scenario_name,
            clear_and_rebuild=False)

        self._journal.generate_report(
            experiment_name=self.active_scenario_name)

    def build_blueprint(
        self,
        type_label: str,
        group_label: str = SCENARIO_GROUP_LABEL
    ) -> None:
        """Triggers a save to disk for a blueprint.

        :param type_label: The type.
        :type type_label: str
        :param group_label: The group. Defaults to SCENARIO_GROUP_LABEL.
        :type group_label: str, optional
        """
        blueprint: FactoryBlueprint = self._factory.get_blueprint(
            group_label=group_label,
            type_label=type_label)


    def register(
        self,
        creator: Any,
        override_group_label: Union[str, None] = None,
        override_type_label: Union[str, None] = None
    ) -> None:
        """Registers an object to the internal factories.

        :param creator: The component to build
        :type creator: Any (implements FactoryImplementable)
        :param override_group_label: Override the group label defined in creator. Defaults to None.
        :type override_group_label: Union[str, None], optional
        :param override_type_label: Override the type label defined in creator. Defaults to None.
        :type override_type_label: Union[str, None], optional
        :param set_default_blueprint: Use the blueprint as the default for this group. Defaults to False.
        :type set_default_blueprint: bool, optional
        """
        type_label: Union[str, None] = None
        group_label: Union[str, None] = None
        # load from component if not overridden

        # group
        if override_group_label is None:
            try:
                group_label = creator.blueprint["group_label"]
            except AttributeError as e:
                print("ERROR: Invalid creator. Creator dump:")
                print(creator)
                print("ERROR Raises:")
                raise e
        else:
            group_label = override_group_label

        # type
        if override_type_label is None:
            type_label = creator.blueprint["type_label"]
        else:
            type_label = override_group_label

        # and if set
        if isinstance(type_label, str) and isinstance(group_label, str):
            try:
                self._load_factories(
                    group_label_override=group_label,
                    type_label_override=type_label,
                    creator=creator)
            except DuplicateFactoryRegistration:
                err_msg = "Duplicate Registration attempted for "\
                    f"{group_label}.{type_label}"

                logging.error(err_msg)
        else:
            msg = "Unable to process blueprint"
            logging.error(msg)
            raise InvalidScenarioState(msg)

    def registry_listing(
        self,
        group_label: Union[str, None] = None
    ) -> list[str]:
        """Lists factory entries. if no group listed it returns groups,
        else if a group is entered it returns all types in that group.

        :param group_label: A group label, defaults to None
        :type group_label: Union[str, None], optional
        :return: A listing of types in a group or all groups
        :rtype: list[str]
        """
        return self._factory.registry_listing(group_label)

    def get_blueprint(
        self,
        type_label: str,
        group_label: str = SCENARIO_GROUP_LABEL
    ) -> FactoryBlueprint:
        """Triggers a save to disk for a blueprint.

        :param type_label: The type.
        :type type_label: str
        :param group_label: The group. Defaults to SCENARIO_GROUP_LABEL.
        :type group_label: str, optional
        """
        return self._factory.get_blueprint(
            group_label=group_label, type_label=type_label)

    def get_description(self, group_label: str, type_label: str) -> str:
        """Gets the description of a component that is registed.

        :param group_label: The group the component is a member of
        :type group_label: str
        :param type_label: the member in the group
        :type type_label: str
        :return: a string of the description as defined in the class.
        :rtype: str
        """
        return self._factory.get_description(
            group_label=group_label,
            type_label=type_label)



"""key tracking imports"""

from .base import MetricTracker

from .structure import (
    AvailableExperimentModel,
    CommentModel,
    ExperimentEntryModel,
    ExperimentRunFolders,
    PrefixModel
)

from .utils import (
    build_folder_listing,
    build_project_folders,
    collect_batch_results,
#    setup_experiment,
    setup_run)



"""Common exceptions for the ADGTK package"""

from typing import Optional

class UnableToMeasureException(Exception):
    """Raise when there is an invalid configuration."""
    def __init__(self, message: Optional[str]=None):
        if message is None:
            message = "Unable to measure the data"
        super().__init__(message)



from .presentation import PresentationFormat
from .records import (
    PresentableRecord,
    SupportsFiltering,
    PresentableGroup)
from .store import (
    SearchOption,
    RecordStore,
    CanExportRecordsToDict,
    CanFindByTerm,
    CanImportRecordsToDict,
    CanFindRandomRecord,
    CanGetAllRecords,
    CanRebuildFromDisk,
    CanSaveToDisk,
    CanSearchForSimilar,
    CanShuffleRecords)



"""structure.py supports the tracking specific data structures


Structures
==========
 - User-facing models: validated, serialized via BaseModel
 - Internal helpers: lightweight dataclasses, not validated

"""

from dataclasses import asdict, dataclass
import datetime
from typing import ClassVar, Optional
import uuid
from pydantic import BaseModel

# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------

EXP_RUN_DIR = "runs"
EXP_OTHER_DIR = "other"
EXP_METRICS_FOLDER = "metrics"
EXP_MODEL_DIR = "models"
EXP_IMG_FOLDER = "images"
EXP_DATASET_FOLDER = "datasets"
EXP_RESULTS_FOLDER = "results"
EXP_MODEL_TRAIN_LOG = "model_train_runs"
TRACKING_FOLDER = ".tracking"

RUN_FOLDER_VERSION = 1.1                # Ease of migration

# ----------------------------------------------------------------------
# Structures - User-facing models (external data)
# ----------------------------------------------------------------------

class CommentModel(BaseModel):
    """Used within the journal for recording comments"""
    comment:str
    component:str = "global"
    timestamp: Optional[str] = None


class ExperimentEntryModel(BaseModel):
    """Used to track experiments within a project."""    
    name:str
    description:str
    journal:str
    results_path:str
    id:Optional[str] = str(uuid.uuid4())
    timestamp: Optional[str] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")


class PrefixModel(BaseModel):
    prefix: str
    major_counter: int = 0
    minor_counter: int = 0


# ----------------------------------------------------------------------
# Structures - Internal helpers (non-validated)
# ----------------------------------------------------------------------

@dataclass
class AvailableExperimentModel():
    name:str
    description: str

@dataclass
class ExperimentRunFolders():
    """Provides a fully qualified path for ease of use for listing the
    different subfolders for an experiment."""
    version: ClassVar[float] = RUN_FOLDER_VERSION
    log_dir:str
    datasets:str
    metrics:str
    images:str
    other: str
    conclusion:str
    root_dir: str
    experiment_name: str
    common: str
    model_dir: str
    train_log_dir:str

    def to_dict(self) -> dict:
        return asdict(self)    



from adgtk.measurements.factory import (
    create_measurement,
    get_measurement_factory_entry,
    get_measurements_by_tag,
    get_measurements_by_type,    
    get_measurement_factory_labels,
    manual_measurement_factory_register,
    register_to_measurement_factory
)
from .engine import MeasurementEngine



from .engine import Impairment, ImpairmentEngine
from .processing import (
    drop_random_value,
    get_longest_string,
    get_random_key,
    already_has_dropped,
    get_random_key_with_number,
    drop_value_for_key
)



"""Managing multiple experiments via CSV using the experiment name as
the key.
"""

import os
import sys
import logging
import time
import csv
from numbers import Number
from typing import Literal, Union, List, Any
# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/folder/.py


# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
#  CSV_DELIMETER = ","
# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------


class ExperimentSubmission:
    column_order: List[str] = ["name", "status", "date_started",
                               "date_completed", "short_description", "run_notes", "observations"]

    def __init__(
        self,
        name: str,
        status: Literal["completed", "canceled", "active", "pending"],
        date_started: Union[float, None] = None,
        date_completed: Union[float, None] = None,
        short_description: Union[str, None] = None,
        run_notes: Union[str, None] = None,
        observations: Union[str, None] = None
    ) -> None:
        if isinstance(name, Number):
            self.name = f"{name}"
        else:
            self.name = name

        self.status = status
        self.run_notes = run_notes
        self.short_description = short_description
        self.observations = observations

        # now generate if empty
        if date_started is None:
            self.date_started = time.time()
        else:
            self.date_started = date_started

        if date_completed is None:
            self.date_completed = time.time()
        else:
            self.date_completed = date_completed

    def get_row(self) -> List[Union[str, int, float, None]]:
        row: List[Union[str, int, float, None]] = []
        for key in self.column_order:
            # required fields
            if key == "name":
                row.append(self.name)
            elif key == "status":
                row.append(self.status)
            elif key == "date_started":
                row.append(self.date_started)
            elif key == "date_completed":
                row.append(self.date_completed)

            # and optional
            elif key == "short_description":
                if self.short_description is not None:
                    row.append(self.short_description)
                else:
                    row.append(" ")
            elif key == "run_notes":
                if self.run_notes is not None:
                    row.append(self.run_notes)
                else:
                    row.append(" ")
            elif key == "observations":
                if self.observations is not None:
                    row.append(self.observations)
                else:
                    row.append(" ")

        return row


# TODO: expand and introduce this class.
class ProjectJournal:
    """Provides a journal for a project. A project is one that extends
    over multiple experiments.
    """

    def __init__(
        self, results_dir: str = "results",
        data_file: str = "experiments.csv",
        overwrite_data_file_if_exists: bool = False
    ):
        self.experiments: dict[str, ExperimentSubmission] = {}
        self.results_dir = results_dir
        self.data_file = os.path.join(self.results_dir, data_file)
        if os.path.exists(self.data_file):
            self._load_data()
        else:
            self._create_file(force=overwrite_data_file_if_exists)

    def __len__(self) -> int:
        return len(self.experiments)

    def _load_data(self) -> None:
        columms: List[str] = []
        with open(self.data_file, "r", encoding="utf-8") as infile:
            csv_reader = csv.reader(infile)

            for row in csv_reader:
                if len(columms) == 0:
                    columms = row

                    # safety check. ensure consistent data.
                    for col in columms:
                        if col not in ExperimentSubmission.column_order:
                            raise KeyError(f"Unexpected column name: {col}")
                    # and the reverse
                    for col in ExperimentSubmission.column_order:
                        if col not in columms:
                            raise KeyError(f"Missing column name: {col}")

                else:
                    data: dict[str, Any] = {}
                    for col, entry in zip(columms, row):
                        data[col] = entry

                    # now create object
                    tmp_exp = ExperimentSubmission(**data)
                    if tmp_exp.name in self.experiments:
                        raise KeyError(
                            f"Duplicate experiment {tmp_exp.name} on load.")
                    self.experiments[tmp_exp.name] = tmp_exp

    def _create_file(self, force: bool = False) -> None:

        # safety check
        if os.path.exists(self.data_file) and not force:
            msg = "Attempted to overwrite the project data file."
            logging.warning(msg)
            sys.exit(os.EX_OSFILE)

        # create the folder if it doesn't exist
        if not os.path.exists(self.results_dir):
            os.makedirs(self.results_dir, exist_ok=True)

        # now the file
        with open(self.data_file, "w", encoding="utf-8") as outfile:
            writer = csv.writer(
                outfile, quoting=csv.QUOTE_MINIMAL)
            writer.writerow(ExperimentSubmission.column_order)

    def _update_disk(self) -> None:
        with open(self.data_file, "w", encoding="utf-8") as outfile:
            writer = csv.writer(
                outfile, quoting=csv.QUOTE_MINIMAL)
            writer.writerow(ExperimentSubmission.column_order)
            for _, item in self.experiments.items():
                writer.writerow(item.get_row())

    def add_entry(
        self,
        name: str,
        status: Literal["completed", "canceled", "active", "pending"],
        date_started: Union[float, None] = None,
        date_completed: Union[float, None] = None,
        short_description: Union[str, None] = None,
        run_notes: Union[str, None] = None,
        observations: Union[str, None] = None,
        overwrite_entry: bool = False
    ) -> None:
        if name in self.experiments and not overwrite_entry:
            logging.warning(f"Unable to add_entry. {name} already found.")
        else:
            self.experiments[name] = ExperimentSubmission(
                name=name,
                status=status,
                date_started=date_started,
                date_completed=date_completed,
                short_description=short_description,
                run_notes=run_notes,
                observations=observations)

        self._update_disk()

    def update_entry(
        self,
        name: str,
        status: Literal["completed", "canceled", "active", "pending"],
        date_started: Union[float, None] = None,
        date_completed: Union[float, None] = None,
        short_description: Union[str, None] = None,
        run_notes: Union[str, None] = None,
        observations: Union[str, None] = None
    ) -> None:
        if name not in self.experiments:
            raise KeyError(f"Experiment {name} not found")

        exp = self.experiments[name]
        if status is not None:
            exp.status = status

        if date_started is not None:
            exp.date_started = date_started

        if date_completed is not None:
            exp.date_completed = date_completed

        if short_description is not None:
            exp.short_description = short_description

        if run_notes is not None:
            exp.run_notes = run_notes

        if observations is not None:
            exp.observations = observations

        self._update_disk()

    def experiment_found(self, name: str) -> bool:
        return name in self.experiments



"""Handlers for the CLI tools for this project
"""

import logging
from typing import Literal, Union
import datetime
import argparse
import sys
import os
import shutil
import signal
import tarfile
import importlib.resources
import http.server
from http.server import HTTPServer
import socketserver
import toml
import yaml
from jinja2 import Environment, FileSystemLoader
from adgtk.common import (
    DEFAULT_SETTINGS,
    DEFAULT_EXP_DEF_DIR,
    DEFAULT_BATCH_DIR)
from adgtk.scenario import ScenarioManager
from adgtk.experiment import ExperimentBuilder
from adgtk.utils import (
    get_user_input,
    start_logging,
    clear_screen,
    create_line,
    load_settings)
from adgtk import __version__ as adgtk_ver
from .intro import intro

# ----------------------------------------------------------------------
# Module configuration
# ----------------------------------------------------------------------
WIZARD_OPTIONS = ("agent", "environment", "policy", "scenario", "custom")

# ----------------------------------------------------------------------
# Module variables
# ----------------------------------------------------------------------
httpd: Union[HTTPServer, None] = None
http_dir = "html"
# ----------------------------------------------------------------------
# Signal handlers
# ----------------------------------------------------------------------


def signal_handler(signum, frame):
    """Handles the signal from the OS

    :param signum: The signal number
    :type signum: int
    :param frame: the frame object
    :type frame: stack frame
    """
    global httpd
    # https://docs.python.org/3/library/signal.html
    if signal.SIGINT == signum:
        print('\nRecievd Ctrl+C! Canceling action.')
        if httpd is not None:
            httpd.server_close()
            print("Server stopped")
    sys.exit(0)

# ----------------------------------------------------------------------
# Web Server
# ----------------------------------------------------------------------


class Handler(http.server.SimpleHTTPRequestHandler):
    """Handler for the web server

    :param http: _description_
    :type http: _type_
    """
    global http_dir

    def __init__(self, *args, directory=http_dir, **kwargs):
        super().__init__(*args, directory=http_dir, **kwargs)


def start_web_server(port: int = 8000, directory: str = "html") -> None:
    """Starts a web server.

    :param port: The port to listen on, defaults to 8000
    :type port: int, optional
    :param directory: The directory to serve, defaults to "html"
    :type directory: str, optional
    """
    global httpd
    global http_dir
    http_dir = directory
    addr = ('', port)
    try:
        # TODO: Fix the type hinting
        httpd = socketserver.TCPServer(addr, Handler)   # type: ignore
        print(f"serving at port {port}")
        if httpd is not None:
            httpd.serve_forever()
        else:
            print("Failed to start server")
            sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)


# ----------------------------------------------------------------------
# Management
# ----------------------------------------------------------------------


def parse_args() -> argparse.Namespace:
    """Parses the command line arguments

    Returns:
        argparse.Namespace: The command line input
    """
    parser = argparse.ArgumentParser()

    # modifiers
    parser.add_argument(
        '-f', '--file',
        type=str,
        help="override the settings file with this file")
    parser.add_argument(
        '--version', action='version', version=f"ADGTK {adgtk_ver}")
    parser.add_argument(
        '--sample',
        action='store_true',
        help="Creates a project with sample code.")
    parser.add_argument(
        "--yaml",
        help="Use YAML format when creating the project settings.file.",
        action="store_true")

    # now positional
    subparsers = parser.add_subparsers(
        dest="command", help="Available commands")

    # Project
    project_parser = subparsers.add_parser("project", help="Project actions")
    project_parser.add_argument(
        "action",
        choices=["create", "destroy"],
        help="Available project actions")
    project_parser.add_argument(
        "name",
        type=str,
        nargs="?",
        help="The project name")

    # Experiment
    exp_parser = subparsers.add_parser(
        "experiment", help="Experiment actions")
    exp_parser.add_argument(
        "action",
        choices=["create", "destroy", "list", "run", "preview", "report"],
        help="Available experiment actions")
    exp_parser.add_argument(
        "name",
        type=str,
        nargs="?",
        help="The experiment name")

    # Factory
    factory_parser = subparsers.add_parser(
        "factory", help="Factory actions")
    factory_parser.add_argument(
        "name",
        type=str,
        nargs="?",
        help="The group name")

    # Batch
    batch_parser = subparsers.add_parser(
        "batch", help="Batch actions")
    batch_parser.add_argument(
        "action",
        choices=["run", "preview", "list", "create"],
        help="Available batch actions")
    batch_parser.add_argument(
        "name",
        type=str,
        nargs="?",
        help="The batch name")

    args = parser.parse_args()
    return args


def build_folders_from_settings():
    """Builds out the folders based on settings"""
    settings = load_settings()
    if not os.path.exists(settings.experiment["data_dir"]):
        os.makedirs(name=settings.experiment["data_dir"], exist_ok=True)
    if not os.path.exists(settings.experiment["results_dir"]):
        os.makedirs(name=settings.experiment["results_dir"], exist_ok=True)
    if not os.path.exists(settings.experiment["tensorboard_dir"]):
        os.makedirs(name=settings.experiment["tensorboard_dir"], exist_ok=True)
    if not os.path.exists(settings.experiment["definition_dir"]):
        os.makedirs(name=settings.experiment["definition_dir"], exist_ok=True)
    if not os.path.exists(settings.logging["log_dir"]):
        os.makedirs(name=settings.logging["log_dir"], exist_ok=True)


def create_settings_file(
    project_name: str,
    data_format: Literal["toml", "yaml"] = "toml",
    clean: bool = False
) -> None:
    """Creates the settings file on project creation. It uses a template
    to create the settings file.

    :param project_name: The project name to be created
    :type project_name: str
    :param data_format: The format of the settings file,
        defaults to "toml"
    :type data_format: Literal[&quot;toml&quot;, &quot;yaml&quot;],
        optional
    :param clean: _description_, defaults to False
    :type clean: bool, optional        
    """

    filename = "project.toml"
    options = "Invalid file type"

    settings_data = DEFAULT_SETTINGS.copy()
    if clean:
        settings_data["user_modules"] = []

    if data_format == "toml":
        options = toml.dumps(settings_data)
    elif data_format == "yaml":
        filename = "project.yaml"
        options = yaml.safe_dump(
            settings_data, sort_keys=False, default_flow_style=False)

    # now craft the file
    env = Environment(loader=FileSystemLoader(
        os.path.join(project_name, 'templates')))
    template = env.get_template("settings.jinja")
    output = template.render(version=adgtk_ver, options=options)

    target_file = os.path.join(project_name, filename)
    with open(file=target_file, encoding="utf-8", mode="w") as outfile:
        outfile.write(output)


def build_experiment(
    experiment_definition_dir: str,
    load_user_modules: list,
    use_formatting: bool = True,
    name: Union[str, None] = None
) -> None:
    """Builds an experiment and saves to disk


    :param name: The name of the experiment, defaults to None
    :type name: Union[str, None], optional
    """
    # Just in case. Build folder(s)
    build_folders_from_settings()

    # and now build the experiment
    builder = ExperimentBuilder(
        experiment_definition_dir=experiment_definition_dir,
        load_user_modules=load_user_modules,
        scenario_manager=None)

    builder.build_interactive(name)

# ---------------------------------------------------------------------
# Project functions
# ---------------------------------------------------------------------


def create_project(
    name: str,
    file_format: Literal["yaml", "toml"] = "toml",
    sample: bool = False
) -> bool:
    """Creates a project

    :param name: The project/directory name to create
    :type name: str
    :param file_format: the format of the settings, defaults to "toml"
    :type file_format: Literal[&quot;yaml&quot;, &quot;toml&quot;], optional
    :param sample: Include sample files, defaults to False
    :type sample: bool, optional
    :return: True if successful, else False
    :rtype: bool
    """
    print(f"Creating project: {sample}")

    if name is None:
        return False

    if os.path.exists(name):
        print("Unable to create a project. remove the other first!")
    elif name is None:
        print("Please specify a name of your project")
    else:
        print(f"Attempting to create {name}")
        if not sample:
            # the clean.tar doesn't have the sample code. See Makefile
            filestream = importlib.resources.files(
                "adgtk").joinpath("clean.tar").open("rb")
        else:
            filestream = importlib.resources.files(
                "adgtk").joinpath("template.tar").open("rb")

        archive = tarfile.open(fileobj=filestream)
        archive.extractall("./")
        archive.close()

        # and rename
        os.rename("project_template", name)

        # safety check
        if os.path.exists(name):
            # now create the settings.py
            clean = True
            if sample:
                clean = False

            create_settings_file(name, data_format=file_format, clean=clean)
            # and let the user know it worked
            print(f"Successfully created project {name}")
            print(f"`cd {name}` to work with the project")
            return True

    print(f"Error creating {name}")
    return False


def destroy_project(name: str) -> bool:
    """Destroys a folder and all its contents

    Args:
        name (str): The folder to destroy
    """
    if name is not None:
        if os.path.exists(name):
            exp_file = os.path.join(name, "project.toml")
            if not os.path.exists(exp_file):
                print("Does not appear to be an ADGTK project. Cancelling")
                return False
            ask_to_remove = f": Please type [{name}] to confirm deleting this project: "
            confirmation = input(ask_to_remove)
            if confirmation.lower() == name.lower():
                shutil.rmtree(name)
                if os.path.exists(name):
                    print("Failed to remove. Please check case")
                else:
                    print(f"Successfully destroyed {name}")
            else:
                print("No action taken.")
            return True
        else:
            print(f"Unable to find project {name} to destroy")
            return True

    return False


def execute(
    experiment: str,
    load_user_modules: list,
    batch_mode: bool = False
) -> None:
    """Runs an experiment

    Args:
        experiment (str): The experiment-definition to run. This is the
            file located in the definition folder as outlined in the
            settings.py file.
    """
    if os.path.exists("project.toml") or os.path.exists("project.yaml"):
        start_logging(name=experiment, surpress_chatty=True, preview=False)
    else:
        print("Execution is unable to find the settings file")
        sys.exit(0)

    loader = ScenarioManager(
        load_user_modules=load_user_modules,
        experiment_definition_dir=DEFAULT_EXP_DEF_DIR)
    loader.load_experiment(experiment)
    if loader.active_scenario is not None:
        loader.run_experiment()
    else:
        msg = "Failure establishing scenario from file"
        logging.error(msg)
        print(msg)
    if not batch_mode:
        sys.exit(0)


def run_batch(batch: str, load_user_modules: list) -> None:
    """Runs a batch of exeperiments

    :param batch: the batch to run
    :type batch: str
    :param load_user_modules: the user modules to load
    :type load_user_modules: list
    """

    if os.path.exists("project.toml") or os.path.exists("project.yaml"):
        start_logging(name=batch, surpress_chatty=True, preview=False)
    else:
        print(
            f"{datetime.datetime.now()} - ERROR: Batch process is unable to "
            "find the settings file")
        sys.exit(0)

    settings = load_settings()
    # create if it doesn't exist
    os.makedirs(settings.batch_dir, exist_ok=True)

    batch_file = os.path.join(settings.batch_dir, f"{batch}.yaml")
    if not os.path.exists(batch_file):
        print(f"Batch file {batch_file} not found")
        sys.exit(0)

    with open(batch_file, "r", encoding="utf-8") as infile:
        batch_data = yaml.safe_load(infile)
        msg = f"{datetime.datetime.now()} - Starting batch {batch} : "\
            f"{batch_data['comments']}"
        print(msg)
        print(create_line(msg, char="="))
        for exp in batch_data["experiments"]:
            print(f"{datetime.datetime.now()} - Starting experiment {exp}")
            execute(
                experiment=exp,
                load_user_modules=load_user_modules,
                batch_mode=True)
            print(create_line(msg, char="-"))
    sys.exit(0)


def preview(experiment: str) -> None:
    """Previews an experiment

    :param experiment: The experiment to preview.
    :type experiment: str
    """
    start_logging(name=experiment, surpress_chatty=True, preview=True)
    loader = ScenarioManager(
        experiment_definition_dir=DEFAULT_EXP_DEF_DIR)
    loader.load_experiment(experiment)
    if loader.active_scenario is not None:
        loader.preview_experiment()
    else:
        print("Failure establishing scenario from file")

    sys.exit(0)


def print_title(clear_screen_first: bool = True) -> None:
    """Clears the screen and prints to console the title

    :param clear_screen_first: Clear screen first, defaults to True
    :type clear_screen_first: bool, optional
    """

    title_string = f"ADGTK - Version {adgtk_ver}"
    line = create_line(title_string)
    if clear_screen_first:
        clear_screen()

    print(line)
    print(title_string)
    print(line)
    print()


# ---------------------------------------------------------------------
# interacting with files and folders
# ---------------------------------------------------------------------
def get_exp_comments_w_basename(file_w_path: str) -> tuple:
    title = os.path.basename(file_w_path)
    comments = "Not set"
    error_found = False
    exp_def = {}
    if file_w_path.lower().endswith(".toml"):
        with open(file_w_path, "r", encoding="utf-8") as infile:
            exp_def = toml.load(infile)
    elif file_w_path.lower().endswith(".yaml"):
        with open(file_w_path, "r", encoding="utf-8") as infile:
            try:
                exp_def = yaml.safe_load(infile)
            except yaml.parser.ParserError as e:
                error_found = True
                comments = f"INVALID: {e}"

    if title.endswith(".toml") or title.endswith(".yaml"):
        title = title[:-5]

    if not error_found:
        if "comments" in exp_def:
            if len(exp_def["comments"]) > 0:
                comments = exp_def["comments"]

    return title, comments


def get_experiment_list(exp_def_dir: Union[str, None] = None) -> list:
    if exp_def_dir is None:
        # try the default
        exp_def_dir = DEFAULT_EXP_DEF_DIR
    try:
        in_files = os.listdir(exp_def_dir)
        files = sorted(in_files)
    except FileNotFoundError:
        print(
            "ERROR: experiment definition directory not found. Check the path.")
        print("Unable to list experiments")
        return []

    return files


def get_exp_to_run(exp_def_dir: Union[str, None] = None) -> str:

    if exp_def_dir is None:
        exp_def_dir = DEFAULT_EXP_DEF_DIR

    experiments = get_experiment_list(exp_def_dir=exp_def_dir)
    if len(experiments) == 0:
        print(f"No experiments found. Please check path for {exp_def_dir}")
        sys.exit(1)

    title = "Experiment"
    comments = "Comments"
    msg = "  Select which experiment to run:"
    line = create_line("", char="-", modified=79)
    print(msg)
    print(line)
    choices = []

    for file in experiments:
        file_w_path = os.path.join(exp_def_dir, file)
        title, comments = get_exp_comments_w_basename(file_w_path)
        choices.append(title)
        print(f"{title:<40} | {comments}")

    print()
    val = get_user_input(
        configuring="experiment",
        request="Which experiment to run?",
        choices=choices,
        requested="str")

    return str(val)


def list_experiments(exp_def_dir: Union[str, None] = None) -> list:

    if exp_def_dir is None:
        exp_def_dir = DEFAULT_EXP_DEF_DIR

    experiments = get_experiment_list(exp_def_dir=exp_def_dir)
    if len(experiments) == 0:
        return []

    title = "Experiment"
    comments = "Comments"
    msg = "  Available experiments"
    line = create_line("", char="-", modified=79)
    print(msg)
    print(line)

    experiments = []

    for file in experiments:
        file_w_path = os.path.join(exp_def_dir, file)
        title, comments = get_exp_comments_w_basename(file_w_path)
        experiments.append(title)
        print(f"{title:<40} | {comments}")

    print(line)
    return experiments


# --------------------------------------------------------------------
# for batch listing
# --------------------------------------------------------------------
def cleanup_name(name: str):
    name = name.replace('.yaml', '')
    name = name.replace('.yml', '')
    name = name.replace('.toml', '')
    return name


def count_results(exp_names: list):
    all_results = os.listdir('results')
    count = 0
    for name in exp_names:
        if name in all_results:
            count += 1
    return count


def list_batch_experiments(batch_dir: str, filename: str) -> tuple:
    batch_file = os.path.join(batch_dir, filename)
    completed_count = 0
    expected_count = 0

    with open(batch_file) as file:
        config = yaml.safe_load(file)
        exp_list = config["experiments"]
        completed_count = count_results(exp_list)
        expected_count = len(exp_list)
        comments = config["comments"]

    return (completed_count, expected_count, comments)


def list_batches(batch_dir: str, show_completed=True) -> list:

    msg = "Available batches"
    line = create_line("", char="-", modified=79)
    print(msg)
    print(line)

    batch_listing = os.listdir(batch_dir)
    batch_names = []
    for batch in batch_listing:
        c_count, e_count, comments = list_batch_experiments(
            batch_dir=batch_dir, filename=batch)
        batch_name = cleanup_name(batch)
        if show_completed:
            batch_names.append(batch_name)
            print(f"- {batch_name:<20} {comments:<60} [{c_count}/{e_count}]")
        elif c_count < e_count:
            batch_names.append(batch_name)
            print(f"- {batch_name:<20} {comments:<60} [{c_count}/{e_count}]")

    return batch_names


def get_batch_to_run(batch_dir: Union[str, None] = None) -> str:

    if batch_dir is None:
        batch_dir = DEFAULT_BATCH_DIR

    batches = list_batches(batch_dir=batch_dir, show_completed=False)
    if len(batches) == 0:
        print(f"No available batches found. Please check path for {batch_dir}")
        sys.exit(1)

    title = "Batch"
    comments = "Comments"
    msg = "  Select which batch to run:"
    line = create_line("", char="-", modified=79)
    print(msg)
    print(line)

    val = get_user_input(
        configuring="batch",
        request="Which batch to run?",
        choices=batches,
        requested="str")

    return str(val)

# --------------------------------------------------------------------
# --------------------------------------------------------------------
#          !!! MANAGER !!! THIS IS THE MAIN FUNCTION
# --------------------------------------------------------------------
# --------------------------------------------------------------------


def manager() -> None:
    """provides a CLI management"""

    signal.signal(signal.SIGINT, signal_handler)

    args = parse_args()
    inside_a_project = False
    # using local variables with a fallback to default for safety

    try:
        if args.file is not None:
            print(f"ADGTK: Loading settings from file {args.file}")
            settings = load_settings(args.file)
        else:
            settings = load_settings()
        exp_def_dir = settings.experiment["definition_dir"]
        batch_def_dir = settings.batch_dir
        user_modules = settings.user_modules
        inside_a_project = True
    except FileNotFoundError:
        exp_def_dir = None
        user_modules = []
    except ValueError:
        msg = f"""
ERROR: Unable to load settings from {args.file}
Reminder: File must end with .toml or .yaml and be properly formatted.
"""
        print(msg)
        sys.exit(1)

    # ----- Project command ------
    if args.command == 'project':
        if args.name is None:
            print("Name of the project is required")
            sys.exit()
        print(args)
        sample = False

        if args.action == "create":
            if args.sample:
                sample = False
                if args.sample:
                    sample = True

            if args.yaml:
                result = create_project(
                    name=args.name, file_format="yaml", sample=sample)
            else:
                result = create_project(
                    name=args.name, file_format="toml", sample=sample)
            if not result:
                print("Error processing\n")
        elif args.action == "destroy":
            if not destroy_project(args.name):
                print("Error processing\n")

        sys.exit(0)
    # ----- batch command ------
    elif args.command == 'batch':
        if not inside_a_project:
            print("No setttings file found. Please check the path")
            sys.exit(0)
        if args.action == "run":
            if args.name is None:
                exp_file = get_batch_to_run(batch_def_dir)
                run_batch(batch=exp_file, load_user_modules=user_modules)
            else:
                # cleanup
                if exp_def_dir in args.name:
                    args.name.replace(exp_def_dir, "")
                    args.name.replace("/", "")
                    args.name.replace("\\", "")
                if args.name.endswith(".toml") or args.name.endswith(".yaml"):
                    args.name = args.run[:-5]

                # and execute
                run_batch(batch=args.name, load_user_modules=user_modules)

        elif args.action == "preview":
            print("Preview not available for a batch. Preview experiments instead")
        elif args.action == "list":
            print_title(clear_screen_first=False)
            _ = list_batches(batch_dir=batch_def_dir, show_completed=True)

        elif args.action == "create":
            print("Batch creation is not supported via the CLI.")

        sys.exit(0)
    # ----- experiment command ------
    elif args.command == 'experiment':
        if not inside_a_project:
            print("No setttings file found. Please check the path")
            sys.exit(0)

        # work the action
        if args.action == "list":
            print_title(clear_screen_first=False)
            _ = list_experiments(exp_def_dir)
        elif args.action == "run":
            if args.name is None:
                exp_file = get_exp_to_run(exp_def_dir)
                execute(experiment=exp_file, load_user_modules=user_modules)
            else:
                # cleanup
                if exp_def_dir in args.name:
                    args.name.replace(exp_def_dir, "")
                    args.name.replace("/", "")
                    args.name.replace("\\", "")
                if args.name.endswith(".toml") or args.name.endswith(".yaml"):
                    args.name = args.run[:-5]

                # and execute
                execute(experiment=args.name, load_user_modules=user_modules)

            sys.exit(0)

        elif args.action == "create":
            print_title(clear_screen_first=True)
            build_experiment(
                experiment_definition_dir=exp_def_dir,
                load_user_modules=user_modules,
                name=args.name)

        elif args.action == "destroy":
            print("Roadmap item. Not implemented yet")
        elif args.action == "preview":
            print_title(clear_screen_first=True)
            preview(args.Preview)
        elif args.action == "report":
            start_web_server(
                port=settings.server["port"],
                directory=settings.experiment["results_dir"])

        sys.exit(0)

    elif args.command == "factory":
        if not inside_a_project:
            print("No setttings file found. Please check the path")
            sys.exit(0)

        print_title(clear_screen_first=False)

        if args.name is not None:
            msg = f"Scenario factory/report for group {args.name}"
            group_label = args.name

            msg = f"Scenario factory/report for group {group_label}"
            line = create_line(msg, char=".")
            print(msg)
            print(line)
            print()
            scenario_mgr = ScenarioManager(load_user_modules=user_modules)
            print(scenario_mgr.group_report(args.name))

            # any problems loading?
            if len(scenario_mgr.load_user_modules_errs) > 0:
                print("ERRORS")
                print("------")
                for err in scenario_mgr.load_user_modules_errs:
                    print(f" - {err}")
        else:
            msg = "Scenario factory/report"
            line = create_line(msg, char=".")
            print(msg)
            print(line)
            print()

            scenario_mgr = ScenarioManager(load_user_modules=user_modules)
            print(scenario_mgr)

            # any problems loading?
            if len(scenario_mgr.load_user_modules_errs) > 0:
                print("ERRORS")
                print("------")
                for err in scenario_mgr.load_user_modules_errs:
                    print(f" - {err}")

        # end of factory option
        sys.exit(0)

    print(intro)
    sys.exit()


if __name__ == '__main__':
    manager()



"""Utilities for handling settings.
"""


import os
from typing import Union
from types import SimpleNamespace
import toml
import yaml

# ----------------------------------------------------------------------
# Module configuration
# ----------------------------------------------------------------------
SETTINGS_FILE = "project.toml"
ALT_SETTINGS_FILE = "project.yaml"

# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------    


def load_settings(file_override: Union[str, None] = None) -> SimpleNamespace:
    """Loads the settings file. Note, giving the user the option to use
    either YAML or TOML.

    Args:
        file_override (Union[str, None], optional): override check for
            settings.[toml|yaml]. Defaults to None.

    Raises:
        FileNotFoundError: File is not found

    Returns:
        SimpleNamespace: settings
    """

    # Note: 
    # Not using configparser due to the desire to support lists. This is
    # needed for allowing the user to list their modules in the settings
    # and have them loaded dynamically.

    
    # Safety checks and set filename to load
    if file_override is not None:
        filename = file_override
        if not os.path.exists(filename):
            raise FileNotFoundError(f"Unable to find {filename}")
        else:
            if filename.endswith(".toml"):
                with open(filename, "r", encoding="utf-8") as infile:
                    values = toml.load(infile)
            elif filename.endswith(".yaml"):
                with open(filename, "r", encoding="utf-8") as infile:
                    values = yaml.safe_load(infile)
            else:
                raise ValueError(
                    "Unsupported or missing file: %s", filename)
    else:
        if os.path.exists(os.path.join(".", SETTINGS_FILE)):
            filename = SETTINGS_FILE
            with open(filename, "r", encoding="utf-8") as infile:
                values = toml.load(infile)
        elif os.path.exists(os.path.join(".", ALT_SETTINGS_FILE)):
            filename = ALT_SETTINGS_FILE  
            with open(filename, "r", encoding="utf-8") as infile:
                values = yaml.safe_load(infile)
        else:
            raise FileNotFoundError("Unable to find settings file")
    
    return SimpleNamespace(**values)



"""For ease of working with the results folder. This module provides
a class to manage the folder structure for an experiment.
"""

import logging
import os
import shutil
import csv
from typing import Union, Any
import yaml
import toml
from adgtk.utils import load_settings

# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/common/test_results.py

# ----------------------------------------------------------------------
# CONSTANTS
# ----------------------------------------------------------------------
DEBUG = False

METRICS_FOLDER = "metrics"
METRICS_IMG_FOLDER = "images"
METRICS_DATA_FOLDER = "data"
DATASET_FOLDER = "datasets"
AGENT_FOLDER = "agent"
PERFORMANCE_FOLDER = "performance"

# ----------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------


def collect_batch_results(exp_prefix: str, results_dir: str) -> list:
    """Retrieves the results and the configuration of an experiment and
    puts it into a dictionary for processing.

    :param experiment_name: The name of the experiment
    :type experiment_name: str
    :return: a tuple of filename and a dict that contains results
    :rtype: dict
    """
    experiments = os.listdir(results_dir)
    results = []
    for experiment in experiments:
        if experiment.startswith(exp_prefix):
            folder_manager = FolderManager(experiment)
            results.append((experiment, folder_manager.collect_results()))

    return results

# ----------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------


class FolderManager:
    """A simple class to help manage and ensure a consistent folder
    structure for an experiment results. It takes the experiment name
    and the settings file to create the necessary folders and
    sub-folders. It also provides an easy way to access the paths to
    the respective use by reading the attributes.

    The class has the following useful attributes:
        * base_folder: the root folder for the experiment
        * agent: the folder for agent data
        * metrics: the folder for metrics
        * metrics_data: the folder for metrics data
        * metrics_img: the folder for metrics images
        * dataset: the folder for datasets
        * performance: the folder for performance data
        * model_dir: the folder for models
    """

    def __init__(
        self,
        name: str,
        settings_file_override: Union[str, None] = None,
        clear_and_rebuild: bool = False
    ):
        """Initialize and create if needed

        :param name: the experiment name
        :type name: str
        :param clear_and_rebuild: deletes all results, defaults to False
        :type clear_and_rebuild: bool
        :param settings_file_override: redirect settings file location,
            defaults to None
        :type settings_file_override: Union[str, None], optional
        :raises FileNotFoundError: Unable to load the settings file
        """

        # cleanup to ensure consistency
        name = name.lower()
        if name.endswith(".toml") or name.endswith(".yaml"):
            name = name[:-5]

        # now load the settings and set the items to manage
        try:
            settings = load_settings(file_override=settings_file_override)

        except FileNotFoundError as e:
            if settings_file_override is not None:
                msg = f"Unable to locate settings file {settings_file_override}"
            else:
                msg = "Unable to locate settings file from default location"
            logging.error(msg)
            raise FileNotFoundError(msg) from e

        exp_result_dir = os.path.join(settings.experiment["results_dir"], name)
        self.base_folder = exp_result_dir

        if os.path.exists(exp_result_dir) and clear_and_rebuild:
            shutil.rmtree(exp_result_dir)

        if not os.path.exists(exp_result_dir):
            os.makedirs(exp_result_dir, exist_ok=True)
            logging.info(f"Created {exp_result_dir}")
        elif DEBUG:
            msg = f"Folder Manager found {exp_result_dir}. No action taken "\
                "to create additional folders"
            logging.info(msg)

        # Agent data
        self.agent = os.path.join(exp_result_dir, AGENT_FOLDER)
        os.makedirs(self.agent, exist_ok=True)

        # Create metrics sub-folder(s)
        self.metrics = os.path.join(exp_result_dir, METRICS_FOLDER)
        self.metrics_data = os.path.join(
            self.metrics, METRICS_DATA_FOLDER)
        self.metrics_img = os.path.join(self.metrics, METRICS_IMG_FOLDER)

        os.makedirs(self.metrics, exist_ok=True)
        os.makedirs(self.metrics_data, exist_ok=True)
        os.makedirs(self.metrics_img, exist_ok=True)

        # create dataset folder
        self.dataset = os.path.join(exp_result_dir, DATASET_FOLDER)
        os.makedirs(self.dataset, exist_ok=True)

        # Performance
        self.performance = os.path.join(exp_result_dir, PERFORMANCE_FOLDER)
        os.makedirs(self.performance, exist_ok=True)

        # Models - added 0.1.1a1
        self.model_dir = "models"
        try:
            self.model_dir = settings.model_dir
        except AttributeError:
            logging.warning("Using older settings file. No model_dir found")
            self.model_dir = os.path.join(exp_result_dir, "models")

        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir, exist_ok=True)

        # and additional/existing files/folders
        self.logfile = os.path.join(settings.logging["log_dir"], f"{name}.log")
        self.data_dir = settings.experiment["data_dir"]
        self.tensorboard_dir = settings.experiment["tensorboard_dir"]
        self.exp_def_dir = settings.experiment["definition_dir"]
        self.name = name

    def __str__(self) -> str:
        """Creates a string representation of the FolderManager object

        :return: a useful string for UX via a CLI.
        :rtype: str
        """
        to_string = "FolderManager\n"
        to_string += "-------------\n"
        to_string += f" - log: {self.logfile}\n"
        to_string += f" - tensorboard: {self.tensorboard_dir}\n"
        to_string += f" - models: {self.model_dir}\n"
        to_string += f" - Performance: {self.performance}\n"
        to_string += f" - agent: {self.agent}\n"
        to_string += f" - metrics: {self.metrics}\n"
        to_string += f"    - data: {self.metrics_data}\n"
        to_string += f"    - images: {self.metrics_img}\n"
        to_string += f" - dataset: {self.dataset}\n"

        return to_string

    def collect_results(self) -> dict:
        """Collects data created through the course of an experiment. It
        does not collect proprietary data such as tensorboard. It also
        does not attempt to read datasets, only the filenames.

        It does read the logfile, performance data, and metrics data.

        :return: The experiment results
        :rtype: dict
        """
        # building blocks
        definition: dict[str, dict] = {}
        performance: dict[str, dict] = {}
        metrics: dict[str, str] = {}
        datasets: list[str] = []

        # assembly
        data: dict[str, Any] = {
            "definition": definition,
            "performance": performance,
            "metrics": metrics,
            "datasets": datasets,
            "logfile": ""
        }

        # experiment definition
        exp_def_yaml = os.path.join(self.exp_def_dir, f"{self.name}.yaml")
        exp_def_toml = os.path.join(self.exp_def_dir, f"{self.name}.toml")

        if os.path.exists(exp_def_yaml):
            with open(exp_def_yaml, "r") as infile:
                data["definition"] = yaml.safe_load(infile)
        elif os.path.exists(exp_def_toml):
            with open(exp_def_toml, "r") as infile:
                data["definition"] = toml.load(infile)
        else:
            data["definition"] = "No experiment definition found"

        # Logfile
        if os.path.exists(self.logfile):
            with open(self.logfile, "r") as infile:
                data["logfile"] = infile.read()
        else:
            data["logfile"] = "No log file found. Possibly part of a batch"

        # performance
        files = os.listdir(self.performance)
        for file in files:
            with open(os.path.join(self.performance, file), "r") as infile:
                # component, label
                "model-train.test_avg_loss.csv"
                component, label, _ = file.split(".")
                if component not in data["performance"]:
                    data["performance"][component] = {}
                csv_reader = csv.reader(infile, quoting=csv.QUOTE_NONNUMERIC)
                for row in csv_reader:
                    # should only be one row based on performance tracker
                    data["performance"][component][label] = row
        # metrics
        files = os.listdir(self.metrics_data)
        for file in files:
            with open(os.path.join(self.metrics_data, file), "r") as infile:
                data["metrics"][file] = infile.read()

        # dataset filenames
        data["datasets"] = os.listdir(self.dataset)

        return data



"""
Used for dealing with formatting and parsing of strings.
"""

import datetime
import yaml
from yaml import YAMLError


# py -m pytest -s test/utils/test_formatting.py


# ----------------------------------------------------------------------
# Module Configurations
# ----------------------------------------------------------------------
DEBUG = False


# ----------------------------------------------------------------------
# Exceptions
# ----------------------------------------------------------------------

class UnableToProcess(Exception):
    """Used for failing to process and wanting to degrade gracefully"""

    default_msg = "Unable to process."

    def __init__(self, message: str = default_msg):
        super().__init__(message)


# ----------------------------------------------------------------------
# Common data representations
# ----------------------------------------------------------------------


def get_timestamp_now(include_time: bool = True) -> str:
    """Generates a string with a timestamp. The goal is to have a
    consistent data and date/time format across the different reports,
    etc.

    Args:
        include_time (bool, optional): Include time?.
            Defaults to True.

    Returns:
        str: A timestamp string formatted for reports, etc.
    """
    now = datetime.datetime.now()
    if include_time:
        return now.strftime("%d/%m/%y at %H:%M:%S")
    else:
        return now.strftime("%d/%m/%y")

# ---------------------------------------------------------------------
# YAML / LLM text processing
# ---------------------------------------------------------------------


def count_prefix_spaces(text: str) -> int:
    cleaned = text
    return len(text) - len(cleaned.strip())


def process_possible_yaml(text: str) -> list:
    """Provides a first attempt at processing YAML. This is ideal when
    the text is properly formatted and a single entry. If direct
    processing fails then this function invokes an attempt to cleanup
    and process that text. This should be invoked as a first attempt for
    text that is likely yaml or at least processable via yaml.

    :param text: the text to process as potential yaml
    :type text: str
    :return: a list of dict that contains the results of processing.
    :rtype: list
    """
    try:
        entry = []
        data = yaml.safe_load(text)
        if data is not None:
            # So its a single entry, convert to list in order to return.
            entry.append(data)
            return entry
    except YAMLError:
        # the root error. uncomment and expand/print as needed below to
        # troubleshoot or learn more.
        # https://pyyaml.org/wiki/PyYAMLDocumentation
        pass
    except yaml.scanner.ScannerError:
        # observed with synthetic data.
        pass
    except yaml.constructor.ConstructorError:
        # can occur if the LLM does something like
        # age: !binary ...
        pass
    except yaml.parser.ParserError:
        pass
    except yaml.composer.ComposerError:
        # may be triggered by ---, observed in the wild
        pass
    except AttributeError:
        pass

    # so we hit a processing error. try again with
    return clean_and_process_possible_yaml(text)


def clean_and_process_possible_yaml(text: str) -> list:
    """cleans up via some simple pre-determined rules and attempts to
    locate and process yaml entries in text. This is useful for when
    there are multiple entries interlaced with text such as when a LLM
    adds text such as "and here is another...", etc. Best practice is to
    run process_possible_yaml and let it invokes this function.

    :param text: the text to process as potential yaml
    :type text: str
    :return: a list of dict that contains the results of processing.
    :rtype: list
    """
    buffer = ""
    results = []
    lines = text.splitlines(keepends=True)
    offset = 0

    # build the buffer
    for line in lines:
        if ":" in line:
            # validate and load the buffer?
            split_out = line.split(":")
            if len(split_out) == 2:
                second_split = split_out[0].split()
                if len(second_split) == 1:
                    # only valid space is a prefix
                    if len(buffer) == 0:
                        # get the offset. This can occur when the output
                        # uses a "tab" to offset text from the examle.
                        # By using an offset the function cleans up this
                        # extra and unexpected spacing.
                        offset = count_prefix_spaces(line)
                        offset -= 1     # so we slice cleanly

                    # process w/offset
                    if offset > 0 and line.startswith(" "):
                        line = line[offset:]

                    # extra formatting needed
                    if line.rstrip().endswith(","):
                        remove = True
                        if line.startswith("'"):
                            remove = False
                        if line.startswith('"'):
                            remove = False

                        if remove:
                            # remove the comma.
                            line = line[:-2]
                            line += "\n"    # and add back in the CR
                    buffer += line
                elif "-" in second_split:
                    # handling of a list object.
                    if len(buffer) > 0:
                        if len(second_split) == 2:
                            buffer += line
        else:
            # process the buffer
            try:
                # safety to confirm buffer is not None and > 3.
                if isinstance(buffer, str):
                    if len(buffer) > 3:
                        data = yaml.safe_load(buffer)
                        if data is not None:
                            results.append(data)
            except YAMLError:
                # the root error. uncomment and expand/print as needed below to
                # troubleshoot or learn more.
                # https://pyyaml.org/wiki/PyYAMLDocumentation
                pass
            except yaml.scanner.ScannerError:
                # observed with synthetic data.
                pass
            except yaml.constructor.ConstructorError:
                # can occur if the LLM does something like
                # age: !binary ...
                pass
            except yaml.parser.ParserError as e:
                pass
                # uncomment to debug and develop. Its normal and
                # expected though to have parse errors with output from
                # a llm though. so this is a NO-OP.
                # raise yaml.parser.ParserError from e
            except yaml.composer.ComposerError:
                # may be triggered by ---, observed in the wild
                pass
            except AttributeError:
                pass

            # and reset buffer if more
            buffer = ""

    return results



"""Builds and manages experiments via a CLI UX

TODO: take advantage of _working_on
"""
# setup logfile for this and sub-modules
from adgtk.utils import create_logger

import os
from typing import Optional, Union
import yaml
from adgtk.utils import get_user_input, get_more_ask
import adgtk.factory.component as factory
from adgtk.factory.structure import BlueprintQuestion
from adgtk.utils import get_user_input
from adgtk.experiment.structure import (
    EXPERIMENT_LABEL,
    SCENARIO_LABEL,
    AttributeEntry,
    ExperimentDefinition)
import adgtk.tracking.project as project_manager
# ----------------------------------------------------------------------
# Module logging
# ----------------------------------------------------------------------

# Set up module-specific logger
_logger = create_logger(
    "adgtk.builder.log",
    logger_name=__name__,
    subdir="framework"
)

# ----------------------------------------------------------------------
# Globals
# ----------------------------------------------------------------------
# public
MIN_LINE_LENGTH = 50

# private
_working_on: list[str] = []        # so we can walk a tree
# ----------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------


def _get_user_selection_from_group(
    group: str,
    user_prompt: str,
    tags: Optional[Union[str, list[str]]] = None,
) -> str:
    """Uses a CLI based experience to list options for the user and
    then via the get_user_input collects the user selection. it then
    returns the associated factory_id of this selection

    :param user_prompt: The prompt to the user
    :type user_prompt: str
    :param group: The group to search for
    :type group: str
    :raises ValueError: Unexpected get_user_input value
    :return: the factory_id of the user selection
    :rtype: str
    """
    entries = factory.list_entries(group=group, tags=tags)

    title = f"Group: {group}:"
    line_length = max(len(title), MIN_LINE_LENGTH)
    print(title)
    print("-"*line_length)
    choices: list[str] = []
    idx_choices = []
    for idx, entry in enumerate(entries):
        print(f"  {idx} : {entry.factory_id:<15} | {entry.summary:50}")
        choices.append(entry.factory_id)
        choices.append(str(idx))
        idx_choices.append(entry.factory_id)
    choice = get_user_input(
        requested="str",
        user_prompt=user_prompt,
        choices=choices,
        allow_whitespace=False)

    # catch numbers.
    # if choice in idx_choices:
    is_index = False
    try:
        _ = int(choice)
        is_index = True
    except ValueError:
        pass

    if is_index:
        # need to convert and return
        if isinstance(choice, str):
            idx = int(choice)
        elif isinstance(choice, int):
            # should really never happen but again, safety
            idx = choice
        else:
            raise ValueError("Unexpected get_user_input value")

        return idx_choices[idx]

    if isinstance(choice, str) and factory.entry_exists(choice):
        return choice

    msg = f"Unexpected get_user_input value: {choice}"
    _logger.error(msg)
    print(f"ERROR: {msg}")
    raise ValueError(msg)


def _perform_interview(
    interview: list[BlueprintQuestion]
) -> list[AttributeEntry]:
    """Performs an interview by iterating through a series of questions.

    :param interview: _description_
    :type interview: list[BlueprintQuestion]
    :raises ValueError: Unexpected combinations
    :return: A list of attributes for init via a factory
    :rtype: list[AttributeEntry]
    """
    item_being_built = _working_on[-1]  # set at _expand

    # not expected but just in case
    if len(interview) == 0:
        raise ValueError("Expand requested with an empty interview")

    attributes = []
    for entry in interview:
        # placeholders
        value: Union[AttributeEntry, str, float, int, bool, list] = ""
        factory_id = None
        if entry.entry_type == "expand":
            if entry.group is None:
                raise ValueError("group must be defined for type expand")
            group_name = entry.group

            if not factory.group_exists(group_name):
                msg = (f"unknown group {group_name}. Valid groups are: "
                       f"{factory.get_group_names()}")
                _logger.error(msg)
                raise ValueError(msg)

            factory_id = _get_user_selection_from_group(
                user_prompt=entry.question,
                group=group_name)
            value = _expand(factory_id=factory_id, attribute=entry.attribute)
        elif entry.entry_type == "ml-string":
            value = get_user_input(
                user_prompt=entry.question,
                requested="ml-str"
            )
        elif entry.entry_type == "list[expand]":
            # get the minimum values
            if entry.group is None:
                raise ValueError("group must be defined for type expand")

            group_name = entry.group
            value = []
            while get_more_ask():
                factory_id = _get_user_selection_from_group(
                    user_prompt=entry.question,
                    group=group_name)

                exp_value = _expand(factory_id=factory_id,
                                    attribute=entry.attribute)
                value.append(exp_value)

        elif entry.choices is not None and len(entry.choices) > 0:
            # handle all choice based calls here
            if entry.entry_type == "bool":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="bool",
                    configuring=item_being_built,
                    choices=entry.choices
                )
            elif entry.entry_type == "str":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="str",
                    configuring=item_being_built,
                    choices=entry.choices
                )
            elif entry.entry_type == "int":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="int",
                    configuring=item_being_built,
                    choices=entry.choices
                )
            elif entry.entry_type == "float":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="float",
                    configuring=item_being_built,
                    choices=entry.choices
                )
            else:
                raise ValueError(
                    "Unexpected requested combination with choices")
        else:
            if entry.entry_type == "bool":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="bool",
                    configuring=item_being_built
                )
            elif entry.entry_type == "str":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="str",
                    configuring=item_being_built
                )
            elif entry.entry_type == "int":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="int",
                    configuring=item_being_built
                )
            elif entry.entry_type == "float":
                value = get_user_input(
                    user_prompt=entry.question,
                    requested="float",
                    configuring=item_being_built
                )
        if isinstance(value, str) and len(value) == 0:
            raise ValueError("Value not captured")

        # now update the list
        if isinstance(value, AttributeEntry):
            attributes.append(value)
        else:
            attributes.append(AttributeEntry(
                attribute=entry.attribute,
                init_config=value,
                factory_id=factory_id
            ))

    return attributes


def _expand(factory_id: str, attribute: str) -> AttributeEntry:
    """Expands an attribute

    :param factory_id: The factory ID to expand
    :type factory_id: str
    :param attribute: The label of the attribute
    :type attribute: str
    :raises ValueError: unknown factory_id
    :return: The collected configuration
    :rtype: AttributeEntry
    """
    global _working_on

    if not factory.entry_exists(factory_id):
        msg = f"Requesting to expand an unknown factory_id {factory_id}"
        _logger.error(msg)
        raise ValueError(msg)
    interview = factory.get_interview(factory_id=factory_id)

    if len(interview) == 0:
        factory_init = False
        init_config: Optional[list] = None
        if attribute == SCENARIO_LABEL:
            # scenarios are unique at the root and must be created
            factory_init = True
            init_config = []

        return AttributeEntry(
            attribute=attribute,
            factory_id=factory_id,
            init_config=init_config,
            factory_init=factory_init
        )

    _working_on.append(factory_id)

    return_value = AttributeEntry(
        factory_id=factory_id,
        attribute=attribute,
        init_config=_perform_interview(interview=interview),
        factory_init=True
    )

    _working_on.pop()
    return return_value


# ----------------------------------------------------------------------
# Public
# ----------------------------------------------------------------------
def create_experiment_name_ux() -> str:
    user_prompt = "Do you want the system to automatically create a name"
    auto_create = get_user_input(
        user_prompt=user_prompt,
        requested="str",
        choices=["yes", "no", "y", "n"],
        allow_whitespace=False,
        default_selection="yes"
    )
    if not isinstance(auto_create, str):
        msg = "create_experiment_name_ux get_user_input failure"
        _logger.error(msg)
        raise ValueError("Unexpected get_user_input")

    if auto_create.lower().startswith("y"):
        cur_prefix_options = project_manager.get_prefix_list()
        if len(cur_prefix_options) > 0:
            print(f"Existing prefixes:")
            for entry in cur_prefix_options:
                print(f" - {entry}")

        helper_str = ("if you want to create a new prefix simply "
                      "enter one here")
        prefix = get_user_input(
            user_prompt="What prefix do you want to use",
            requested="str",
            allow_whitespace=False,
            helper=helper_str,
            min_characters=1,
            max_characters=20)
        if not isinstance(prefix, str):
            msg = "create_experiment_name_ux get_user_input failure"
            _logger.error(msg)
            raise ValueError("Unexpected get_user_input")
        # return here
        return project_manager.generate_experiment_name(
            prefix=prefix, update_next="minor")

    name = get_user_input(
        user_prompt="Please enter an experiment name",
        requested="str",
        allow_whitespace=False)
    if not isinstance(name, str):
        msg = "create_experiment_name_ux get_user_input failure"
        _logger.error(msg)
        raise ValueError("Unexpected get_user_input")

    return name


# TODO: add tags, etc.
def build_experiment(
    name: Optional[str] = None,
    scenario_factory_id: Optional[str] = None,
    tags: Optional[str] = None
) -> None:

    if name is None:
        name = create_experiment_name_ux()

    description = get_user_input(
        user_prompt="Experiment description",
        requested="str",
        max_characters=80,
        min_characters=1
    )

    if scenario_factory_id is not None:
        # check first
        if not factory.entry_exists(scenario_factory_id):
            print("WARNING: Scenario with that factory_id does not exist.")
        scenario_factory_id = None

    if scenario_factory_id is None:
        scenario_factory_id = _get_user_selection_from_group(
            group="scenario",
            user_prompt="Please select the scenario you wish to build",
            tags=tags)

    scenario_config = _expand(scenario_factory_id, attribute=SCENARIO_LABEL)

    if not isinstance(name, str):
        raise ValueError("Unexpected return for value")
    if not isinstance(description, str):
        raise ValueError("Unexpected return for description")

    exp_def = ExperimentDefinition(
        name=name,
        attribute=EXPERIMENT_LABEL,
        description=description,
        init_config=scenario_config,
        factory_init=True
    )

    filename = f"{name}.yaml"
    file_w_path = os.path.join("blueprints", filename)
    with open(file=file_w_path, mode="w", encoding="utf-8") as outfile:
        yaml.safe_dump(
            exp_def.model_dump(),
            outfile,
            default_flow_style=False,
            sort_keys=False)

    _logger.info(f"Created experiment file: {file_w_path}")



"""File operations utility functions


"""

import logging
import csv
import json
import os
import pickle
import random
from typing import cast, Any, Iterable, Optional, Union
import pandas as pd
from pydantic import ValidationError
from adgtk.data.structure import(
    OrientationTypes,
    FileDataDefinition,
    FileDefinition,
    InMemoryDataDefinition
)

# ----------------------------------------------------------------------
# Constants and types
# ----------------------------------------------------------------------

ReturnDataTypes = Union[dict, list, pd.DataFrame, None]

# ----------------------------------------------------------------------
# Inspection Functions
# ----------------------------------------------------------------------

def valid_dict_of_lists(data:dict) -> bool:
    """Verifies a dict is valid for actions such as shuffling/flipping.

    :param data: The data to inspect
    :type data: dict
    :return: True if a candidate for shuffling or flipping
    :rtype: bool
    """
    keys = list(data.keys())
    list_lengths = [len(data[k]) for k in keys]
    
    if len(set(list_lengths)) != 1:
        return False
    return True


def inspect_current_orientation(data:ReturnDataTypes) -> OrientationTypes:
    """Based on the data returns the orientation type.

    :param data: The data to inspect
    :type data: ReturnDataTypes
    :return: the orientation of the data
    :rtype: OrientationTypes
    """
    found_orientation = "other"
    if isinstance(data, pd.DataFrame):
        found_orientation = "pandas"
    elif isinstance(data, list):
        if isinstance(data[0], dict):
            found_orientation = "list_contains_dict"
        elif isinstance(data[0], str):
            found_orientation = "list_contains_string"
    elif isinstance(data, dict):
        # assumption:
        found_orientation = "dict_contains_list"
        for key, value in data.items():
            if not isinstance(value, list):
                # reduced precision
                found_orientation = "dict"      
    # pylance/tox is complaining about str instead of literal
    return found_orientation        # type: ignore

# ----------------------------------------------------------------------
# Loading Functions
# ----------------------------------------------------------------------

def load_data_from_csv_file(filename:str) -> list:
      """Loads a CSV file into a list of dictionaries.

      :param filename: The name of the file to load
      :type filename: str
      :return: The data from the file
      :rtype: list
      """
      columns: list[str] = []
      records: list[dict] = []

      with open(filename, "r") as infile:
            csv_reader = csv.reader(infile)
            for row in csv_reader:
                  # we are on the first row when len == 0
                  if len(columns) == 0:
                        columns = row
                  else:
                        data: dict[Any, Any] = {}
                        for idx, col in enumerate(columns):
                              data[col] = row[idx]
                        records.append(data)
      return records


def load_data_from_file(
    file_def: FileDefinition
) -> ReturnDataTypes:
    """Loads data from file

    :param file_def: _description_
    :type file_def: FileDefinition
    :raises ValueError: Unexpected value
    :return: _description_
    :rtype: ReturnDataTypes
    """
    file_w_path = os.path.join(file_def.path, file_def.filename)
    encoding = file_def.encoding

    # --------- load the data from file ---------
    # First, load the data for the types that can transform
    df:Optional[pd.DataFrame] = None
    data:Optional[Union[pd.DataFrame, dict, list]] = None

    if encoding == "csv":
        data = load_data_from_csv_file(file_w_path)

    elif encoding == "json":
        with open(file_w_path, encoding="utf-8", mode="r") as infile:
             data = json.load(infile)
    elif encoding == "pickle":
        with open(file_w_path, mode="rb") as infile:
            data = pickle.load(infile)
    elif encoding == "pandas":
        if file_w_path.endswith(".csv"):
            data = pd.read_csv(file_w_path)
        elif file_w_path.endswith(".pkl"):
            data = pd.read_pickle(file_w_path)
        else:
            raise ValueError("Unexpected file extension")
    else:        
        logging.warning(
            f"Unknown encoding defined for {file_def}")
        raise ValueError("Unsupported encoding")

    if data is None:
        ValueError("Data is None after load.")
    return data

def change_orientation(
    data:ReturnDataTypes,
    target_orientation: OrientationTypes
) -> ReturnDataTypes:
    """This function changes the orientation of data to align with a
    target. 

    :param data: The data to transform
    :type data: _type_
    :param orientation: The target data format
    :type orientation: OrientationTypes
    :raises RuntimeError: failed to transform

    :return: The data modified into the new orientation
    :rtype: ReturnDataTypes
    """

    # ------- identify source orientation ------
    found_orientation = inspect_current_orientation(data)  
    if found_orientation == target_orientation:
        return data

    # ------- target transformations -------
    if found_orientation == "list_contains_string":
        if target_orientation == "list_contains_string":
            return data
    elif found_orientation == "pandas":
        # only way here is if found is pandas
        data = cast(pd.DataFrame, data)

        if target_orientation == "pandas":
            return data
        elif target_orientation == "list_contains_dict":                
            return flip_from_dict_to_list(data.to_dict())            
        elif target_orientation == "dict_contains_list":
            return data.to_dict()
        else:
            logging.warning(f"Unexpected target_orientation {target_orientation} for pandas source")
            return data.to_dict()
    elif found_orientation == "list_contains_dict":
        data = cast(list, dict)
        if target_orientation == "list_contains_dict":
            return data
        elif target_orientation =="dict_contains_list":
            return flip_from_list_to_dict(data)
        elif target_orientation == "pandas":
            return pd.DataFrame(data)
    elif found_orientation == "dict_contains_list":
        data = cast(dict, data)
        if target_orientation == "dict_contains_list":
            return data        
        elif target_orientation == "list_contains_dict":
            return flip_from_dict_to_list(data)        

    raise RuntimeError("Failed to transform.")


# ----------------------------------------------------------------------
# Transformation functions
# ----------------------------------------------------------------------


def flip_from_list_to_dict(data:list) -> dict:
    """Flips a list of data into a dict

    :param data: The data to transform
    :type data: list
    :raises ValueError: No data, malformed data for flipping
    :raises RuntimeError: general error with method
    :return: a list transformed into a dict with lists
    :rtype: dict
    """

    if len(data) == 0:
        raise ValueError("No data to flip")
    
    # initialize
    flipped:dict[str,list] = {}

    # process
    for row in data:
        try:
             for key, value in row.items():
                if key not in flipped.keys():
                    # add a list
                    flipped[key] = []
                # now append
                flipped[key].append(value)
        except IndexError:
             raise ValueError("Unable to flip_from_list_to_dict")
    # safety check
    if valid_dict_of_lists(flipped):
        return flipped
    
    raise RuntimeError("Error flipping data from list to dict")

def flip_from_dict_to_list(data:dict) -> list:
    """Converts a dict of lists into a list of dicts

    :param data: The data to transform
    :type data: dict
    :raises ValueError: Data is not properly formatted for flipping.
    :return: A list of dicts containing the same data elements
    :rtype: list
    """
    dest = []
    if not valid_dict_of_lists(data):
        msg = "Invalid dict type for flipping. Lists are different lengths"
        raise ValueError(msg)
    
    keys = list(data.keys())
    
    for idx in range(len(data[keys[0]])):
        to_insert = {}
        for key in keys:
            to_insert[key] = data[key][idx]
        dest.append(to_insert)
    return dest
    

def remap_data(
    data:ReturnDataTypes,
    key_map:dict[str,str]
) -> ReturnDataTypes:
    """Remaps data keys

    :param data: The data to remap
    :type data: ReturnDataTypes
    :param key_map: The mapping {from:to}
    :type key_map: dict[str,str]
    :raises TypeError: Unknown data type
    :return: The updated data with the new keys/columns
    :rtype: ReturnDataTypes
    """

    if isinstance(data, pd.DataFrame):
        return data.rename(columns=key_map)
    elif isinstance(data, dict):
        for old_key, new_key in key_map.items():
            if old_key in data.keys():
                data[new_key] = data.pop(old_key)
    elif isinstance(data, Iterable):
         for row in data:
              for old_key, new_key in key_map.items():
                   if old_key in row.keys():
                        row[new_key] = row.pop(old_key)
    else:
        raise TypeError(f"Unexpected Data type {type(data)} for remap_data")
    
    return data


def shuffle_dict_of_lists(data: dict) -> dict:
    """Shuffles a dict containing lists

    :param data: The data to shuffle
    :type data: dict
    :raises ValueError: Unexpected list length within dict
    :return: A shuffled dict of lists
    :rtype: dict
    """

    if not valid_dict_of_lists(data):
        msg = "Invalid dict type for shuffling. Lists are different lengths"
        raise ValueError(msg)
    
    keys = list(data.keys())
    list_lengths = [len(data[k]) for k in keys]
    indices = list(range(list_lengths[0]))
    random.shuffle(indices)

    return {k: [data[k][i] for i in indices] for k in keys}


def shuffle_data(data:ReturnDataTypes) -> ReturnDataTypes:
    """Shuffles the data by inspecting then shuffling based on the
    data type found.

    :param data: The data to shuffle
    :type data: ReturnDataTypes
    :return: The shuffled data
    :rtype: ReturnDataTypes
    """
    if isinstance(data, list):
        random.shuffle(data)
        return data # random.shuffle is in-place but return=consistent
    elif isinstance(data, dict):
        return shuffle_dict_of_lists(data)
    elif isinstance(data, pd.DataFrame):
        return data.sample(frac=1).reset_index(drop=True)
    raise ValueError("Unexpected data type")

def split_dict(data:dict, keys:list) -> tuple:
    """Splits a dictionary into left/right based on keys. If a string is
    the the key in the data element and in the keys list then it is sent
    to the right record, else the left. This allows therefore an easy to
    use function for splitting left/right records. Note, this works
    regardless of what the values are, float, str, list, etc.

    :param data: The data to split.
    :type data: dict
    :param keys: a list of strings that mark what maps to right
    :type keys: list
    :return: two dicts (left, right)
    :rtype: tuple
    """
    left = {}
    right = {}

    for key, value in data.items():
        if key in keys:
            right[key] = value
        else:
            left[key] = value
    return left, right


def split_data_into_left_right(
    data: Union[list, dict, pd.DataFrame],
    keys:list
) -> tuple:
    """Splits data into left/right elements.

    :param data: The data to split
    :type data: Union[list, dict, pd.DataFrame]
    :param keys: a list of strings that mark what maps to right
    :type keys: list
    :return: a left and right of the same type as before
    :rtype: tuple
    """
     
    if isinstance(data, dict):
        return split_dict(data=data, keys=keys)
    elif isinstance(data, pd.DataFrame):
        left = data.drop(columns=keys, errors="ignore")
        right = data[keys].copy()
    elif isinstance(data, list):
        left = []
        right = []
        for row in data:
            if isinstance(row, dict):
                l_row, r_row = split_dict(data=row, keys=keys)
                left.append(l_row)
                right.append(r_row)
    return left, right


# ----------------------------------------------------------------------
# Orchestration functions
# ----------------------------------------------------------------------

def load_data(
    data_def: Union[
        InMemoryDataDefinition,
        FileDataDefinition,
        FileDefinition,
        dict]
) -> ReturnDataTypes:
    """Loads the data and performs the requested data wrangling such as
    remapping of keys, format, and shuffle.

    :param data_def: The definition to load based on
    :type data_def: Union[
        InMemoryDataDefinition, FileDataDefinition. FileDefinition, dict]
    :raises ValueError: Invalid data_def.
    :return: The loaded and transformed data
    :rtype: ReturnDataTypes
    """
    
    # --------- loading of data -----------
    data = None
    if isinstance(data_def, dict):
        # Try FileDataDefinition
        try:
            data_def = FileDataDefinition(**data_def)
        except ValidationError:
            pass
        # file def not file data def?
        try:
            data_def = FileDefinition(**data_def)   # type: ignore

        except ValidationError:
            pass            
        # inMemory?
        try:
            data_def = InMemoryDataDefinition(**data_def)   # type: ignore
        except ValidationError:
            raise ValueError("Unexpected data definition. Unable to load")

    # now processing
    if isinstance(data_def, FileDataDefinition):
        data = load_data_from_file(file_def=data_def.file_definition)
    elif isinstance(data_def, InMemoryDataDefinition):
        data = data_def.data
    elif isinstance(data_def, FileDefinition):
        data_def = FileDataDefinition(
            shuffle_on_load=False,
            key_rename_map=None,
            target_orientation=None,
            file_definition=data_def
        )
        load_data_from_file(file_def=data_def.file_definition)

    # confirm load? None is acceptable to return.
    if data is None:
        return None


    # --------- processing of data -----------
    if data_def.shuffle_on_load is not None:
        if data_def.shuffle_on_load:
            data = shuffle_data(data)
    
    # change orientation?
    target_orientation = data_def.target_orientation
    if target_orientation is not None:
        data = change_orientation(
            data=data, target_orientation=target_orientation)

    # remapping requested?
    if data_def.key_rename_map is not None:
        data = remap_data(data=data, key_map=data_def.key_rename_map)

    # and return
    return data



"""Structure for builder

Structures
==========
 - User-facing models: validated, serialized via BaseModel
 - Internal helpers: lightweight dataclasses, not validated
"""

from typing import (
    Any,
    Callable,
    Optional,
    Protocol,
    Union,
    get_args,
    runtime_checkable)
from pydantic import BaseModel
from adgtk.data.structure import FileEntry
from adgtk.tracking.structure import ExperimentRunFolders

# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------

EXPERIMENT_LABEL = "experiment"
SCENARIO_LABEL = "scenario"

AttributeValueType = Union["AttributeEntry", dict, list, str, bool, float, int]


# ----------------------------------------------------------------------
# Structures - User-facing models (external data)
# ----------------------------------------------------------------------

class AttributeEntry(BaseModel):
    """Enables the runner to know how to load the data."""
    attribute: str    
    factory_id: Optional[str] = None
    factory_init: bool = False
    init_config: Optional[
        Union[
            bool,
            int,
            float,
            str,
            "AttributeEntry",
            dict,
            list["AttributeEntry"]]
            ] = None
    
AttributeEntry.model_rebuild()  # str to Type AttributeEntry for Union


class ExperimentDefinition(AttributeEntry):
    """The root of any experiment has two extra items the name and
    description. This allows for easy management and prediction of what
    to expect on load."""
    name:str
    description: str

class BatchDefinition(BaseModel):
    name:str
    experiments: list[str]

class ScenarioResults(BaseModel):
    """The results of an experiment"""
    files: list[FileEntry]


# ----------------------------------------------------------------------
# Structures - Internal helpers (non-validated)
# ----------------------------------------------------------------------


# ----------------------------------------------------------------------
# Protocols / Abstract Classes
# ----------------------------------------------------------------------

@runtime_checkable
class ScenarioProtocol(Protocol):
    """Defines a scenario"""
    def run_scenario(
        self,
        result_folders:ExperimentRunFolders
    ) -> ScenarioResults: # type: ignore
        """Runs the scenario as defined"""


# ----------------------------------------------------------------------
# Typing
# ----------------------------------------------------------------------

BuildComponentResult = Union[Callable, float, bool, str, int, list[Any], None]
AttributeConfigType = Union[AttributeEntry, dict]

ATTRIBUTE_CONFIG_TYPES = tuple(get_args(AttributeConfigType))



"""data.tracking.py is focused on providing a re-usable class based on
data usage needs. The overall goal of the data module to which the
tracking is the core is to provide a consistent and repeatable method
for letting agents know what files/data are available and for what
purpose.

Roadmap
=======
1. consider YAML and sqlite3 as additional file tracking solutions.
2. refactor report for better UI experience.
"""
import logging
import json
import os
from typing import Optional, Union
import uuid
from pydantic import ValidationError
from adgtk.data.structure import FileDefinition, FileEncodingTypes


# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------

class JsonFileTracker:
    """A simple file tracker that implements the CanTrackFiles protocol.
    It uses a json file the user defines to keep an inventory of files
    that can be used/referenced/loaded.
    """

    def __init__(
        self,
        label: str,
        inventory_file: str,
        logger: Optional[logging.Logger] = None
    ) -> None:
        self.label = label
        self.inventory_file = inventory_file
        self._inventory = {}
        self.logger = logger or logging.getLogger(__name__)        
        self._load_from_disk()

    def _load_from_disk(self) -> None:
        """Loads the inventory from disk. If it doesn't exist it creates it
        in memory for a future save. The method expects JSON format.
        """
        if os.path.exists(self.inventory_file):
            with open(self.inventory_file, "r", encoding="utf-8") as infile:
                self._inventory = json.load(infile)

                for id, entry in self._inventory.items():
                    if not isinstance(entry, FileDefinition):
                        try:
                            entry = FileDefinition(**entry)
                            self._inventory[id] = entry
                        except ValidationError:
                            msg = ("Potential corruption when loading from "
                                   f"disk : with file {self.inventory_file}")
                            self.logger.error(msg)
                            raise Exception(msg)
                msg = (f"Loaded {len(self._inventory)} into "
                       "JsonFileTracker from disk")
                self.logger.info(msg)

        else:
            self.logger.warning(
                f"unable to load {self.inventory_file}. This is expected if "
                "creating a new tracker.")
            self._inventory = {}

    def _save_to_disk(self) -> None:
        """Saves the inventory to disk in JSON format."""
        with open(self.inventory_file, "w", encoding="utf-8") as outfile:
            out_data = {}
            for id, entry in self._inventory.items():
                if isinstance(entry, FileDefinition):
                    out_data[id] = entry.model_dump()
                else:
                    out_data[id] = entry
            json.dump(out_data, outfile, indent=2)
        self.logger.info(
            f"{self.label} saved inventory to {self.inventory_file}")

    def list_files(
        self,
        tag: Optional[Union[str, list]] = None
    ) -> list[FileDefinition]:
        """Provides a listing of files with the ability to filter based
        on tags.

        :param tag: The tag(s) to filter for, defaults to None
        :type tag: Optional[Union[str, list]], optional
        :raises ValueError: Corrupt inventory
        :return: A list of file definitions that match the query
        :rtype: list[FileDefinition]
        """

        found: list[FileDefinition] = []
        file: FileDefinition
        for _, file in self._inventory.items():
            # ensure consistent format
            if isinstance(file, FileDefinition):
                pass
            else:
                try:
                    file = FileDefinition(**file)
                except ValidationError:
                    raise ValueError("Corrupted inventory")
            # inspect each file. not fast but it works
            if tag is None:
                found.append(file)
            elif file.tags is not None:
                if isinstance(tag, list):
                    if all(entry in file.tags for entry in tag):
                        found.append(file)
                elif isinstance(tag, str):
                    if tag in file.tags:
                        found.append(file)
        return found
    
    def get_file_ids_only(
        self,
        tag: Optional[Union[str, list]] = None
    ) -> list[str]:
        files = self.list_files(tag=tag)
        return [file.file_id for file in files]


    def get_file_id(self, filename: str, path: Optional[str] = None) -> str:
        """Retrieves the file_id for the file requested.

        :param filename: The filename to search for
        :type filename: str
        :param path: The path override, defaults to None
        :type path: Optional[str], optional
        :raises ValueError: Corrupted inventory
        :raises FileNotFoundError: No file found for the filename
        :return: the file_id of the entry
        :rtype: str
        """

        file: FileDefinition
        for _, file in self._inventory.items():
            # ensure consistent format
            if isinstance(file, FileDefinition):
                pass
            else:
                try:
                    file = FileDefinition(**file)
                except ValidationError:
                    raise ValueError("Corrupted inventory")

            if file.filename == filename and file.path == path:
                return file.file_id

        raise FileNotFoundError()

    # TODO: refactor to provide better UI
    def report(self, tag: Optional[Union[str, list]] = None) -> None:
        """Generates a report and prints to console of all the files that
        are curently in the inventory.

        :param tag: The tag(s) to filter for, defaults to None
        :type tag: Optional[Union[str, list]], optional
        """

        longest = 0
        all_files = ""
        files = self.list_files(tag=tag)
        files.sort(key=lambda x: (x.path, x.filename))
        for file in files:
            path = file.path

            if file.tags is None:
                tags = ""
            else:
                tags = " ".join(file.tags)
            entry = (f" - {file.file_id:<35} | {file.filename:<30} | "
                     f"{path:<15} | {tags}")
            if len(entry) > longest:
                longest = len(entry)

            all_files += f"{entry}\n"

        # Setup title/'banner', the spaced out title for the columns
        title = f"{self.label} File Manager report"
        if tag is not None:
            if isinstance(tag, str) and len(tag) > 0:
                title = f"{self.label} File Manager report - tag: {tag}"
            elif isinstance(tag, list) and len(tag) > 0:
                tag_str = " ".join(tag)
                title = f"{self.label} File Manager report - tags: {tag_str}"

        filename = "Filename"
        path_str = "Folder"
        file_id_str = "File ID"
        banner = (f"    {file_id_str:<35} | {filename:<30} | "
                  f"{path_str:<15} | Tags")
        if len(title) > longest:
            longest = len(title)
        if len(banner) > longest:
            longest = len(banner)

        if longest > len(title):
            # now center the title
            spaces = int((longest-len(title))/2)
            space_str = " "*spaces
            title = f"\n{space_str}{title}"
        # and finally put everything together and print
        line = "="*longest
        small_line = "-"*longest
        title += f"\n{line}\n{banner}\n{small_line}\n{all_files}"
        print(title)

    def register_file(
        self,
        source_file: str,
        encoding: FileEncodingTypes,
        metadata_file: Optional[str] = None,
        tags: Optional[Union[str, list[str]]] = None,
        id: Optional[str] = None
    ) -> str:
        """Registers a file and if requested moves to internal data folder

        :param source_file: The name of the file w/path
        :type source_file: str
        :param encoding: The encoding of the file
        :type encoding: FileEncodingTypes
        :param tags: The tags for this file, defaults to None
        :type tags: Optional[list[str]], optional
        :param id: The requested ID, defaults to None
        :type id: Optional[str], optional
        :raises FileNotFoundError: File is not found
        :raises IndexError: The id already exists
        :return: The id of the file saved
        :rtype: str
        """

        # Do we need to move the file?
        if not os.path.exists(source_file):
            raise FileNotFoundError(f"File must exist on disk: {source_file}.")

        # create an id if not already found
        if id is None:
            id = str(uuid.uuid4())

        # safety
        if id in self._inventory.keys():
            raise IndexError(f"ID: {id} already exists")

        # and prepare tags
        if tags is None:
            tags = []

        # Split into file and path
        dir, filename = os.path.split(source_file)

        # now create entry
        entry = FileDefinition(
            file_id=id,
            filename=filename,
            path=dir,
            encoding=encoding,
            metadata_file=metadata_file,
            tags=tags
        )

        self._inventory[id] = entry
        self._save_to_disk()
        self.logger.info(
            f"{self.label} created entry: {id} for file {source_file}")
        return id

    def retire_file(self, id: str) -> None:
        """Removes a file from the inventory. If the file is stored in the
        internal directory the file is moved to the .trash folder.

        :param id: The id of the file to retire
        :type id: str
        :raises IndexError: Unknown ID
        :raises InvalidConfiguration: Corrupted data
        :return: True if file retired
        :rtype: bool
        """
        if id not in self._inventory.keys():
            raise IndexError(f"Unknown ID: {id}")

        # now delete the entry
        del self._inventory[id]
        self.logger.info(
            f"{self.label} retired entry: {id}")
        self._save_to_disk()

    def get_file_definition(self, id: str) -> FileDefinition:
        """Retrieves the file definition for an ID.

        :param id: The id of the file
        :type id: str
        :raises KeyError: Unable to fine the file ID
        :return: The details associated with this file.
        :rtype: FileDefinition
        """

        if id in self._inventory.keys():
            return self._inventory[id].model_copy()

        msg = f"Unable to find file id {id}"
        raise KeyError(msg)

    def file_id_exists(self, id:str) -> bool:
        """An easy method to verify an id exists.

        :param id: The ID to check
        :type id: str
        :return: True if a valid ID in the system.
        :rtype: bool
        """

        if id in self._inventory.keys():
            return True
        return False


"""The Journal collects data about an experiment and creates a markdown
document in the results folder with key information and other details to
return to experiment results in the future.
"""


# py -m pytest test/journals/test_journal.py
import logging
import os
from urllib.parse import urljoin
from typing import Literal, Union
import jinja2
from jinja2 import Environment, FileSystemLoader
from adgtk.utils import get_timestamp_now
from .base import SupportsReportingOperations
from adgtk.common import DEFAULT_JOURNAL_REPORTS_DIR, FolderManager
from adgtk.utils import load_settings


# ----------------------------------------------------------------------
# Module configuration
# ----------------------------------------------------------------------
PREVIEW_REPORT_FILENAME = "preview.html"
REPORT_FILENAME = "report.html"

# ----------------------------------------------------------------------
# Journal
# ----------------------------------------------------------------------


class ExperimentJournal:
    """Central place for different components to register comments, etc.
    Will also be responsible for creating reports using Jinja2 templates
    that are in the users template directory. This way the user is free
    to update CSS or similar from the template."""

    def __init__(
        self,
        use_formatting: bool = True,
        settings_file_override: Union[str, None] = None
    ) -> None:
        """Creates a new instance of the ExperimentJournal.

        :param use_formatting: staging for CSS, defaults to True
        :type use_formatting: bool, optional
        :param settings_file_override: a settings file, defaults to None
        :type settings_file_override: Union[str, None], optional
        """
        self.settings_file_override = settings_file_override
        self.use_formatting = use_formatting
        self._scenario_def: str = "NOT-SET"
        self._comments: list[str] = []
        self._measurements: list[str] = []
        self._data_created: dict[str, list] = {}
        self._data_sample: dict[str, list] = {}
        self._data_other: dict[str, list] = {}
        self._data_metrics: dict[str, list] = {}
        self._data_global: list[str] = []
        self._tools: list[str] = []

       # now load the settings and set the items to manage
        try:
            settings = load_settings(file_override=settings_file_override)
        except FileNotFoundError as e:
            if settings_file_override is not None:
                msg = f"Unable to locate settings file {settings_file_override}"
            else:
                msg = "Unable to locate settings file from default location"
            logging.error(msg)
            raise FileNotFoundError(msg) from e

        # public and can be read/written but not worth cluttering params
        # these should not change often and likely only in special
        # cases such as dynamic sub-experiments, etc.
        self.report_folder = DEFAULT_JOURNAL_REPORTS_DIR
        self.preview_report_filename = PREVIEW_REPORT_FILENAME
        self.report_filename = REPORT_FILENAME
        self.engines: list[SupportsReportingOperations] = []
        server_proto = settings.server["proto"]
        server_host = settings.server["host"]
        server_port = settings.server["port"]
        self.base_url = f"{server_proto}://{server_host}:{server_port}"

    def register_meas_engine(self, engine: SupportsReportingOperations) -> None:
        """Registers a measurement engine for reporting purposes

        :param engine: The engine
        :type engine: SupportsReportingOperations
        """
        logging.info(f"Journal registered {engine.name}")
        self.engines.append(engine)

    def reset(self) -> None:
        """clears all data from the internal structure.
        """
        self._scenario_def = "NOT-SET"
        self._comments = []
        self._measurements = []
        self._data_created = {}
        self._data_metrics = {}
        self._data_other = {}
        self._data_sample = {}
        self._data_global = []

    def generate_preview(
        self,
        experiment_name: str,
        experiment_folder: str
    ) -> None:
        """Generates a preview of an experiment and writes that
        preview to disk.

        :param experiment_name: The name of the experiments
        :type experiment_name: str
        :param experiment_folder: root folder for the experiment results
        :type experiment_folder: str
        """
        report_folder = os.path.join(experiment_folder, self.report_folder)

        if not os.path.exists(report_folder):
            os.makedirs(report_folder, exist_ok=True)

        report_filename = os.path.join(
            report_folder, self.preview_report_filename)

        env = Environment(loader=FileSystemLoader('templates'))
        template = env.get_template("preview.jinja")

        try:
            output = template.render(
                experiment_name=experiment_name,
                comments=self._comments,
                measurements=self._measurements,
                scenario_def=self._scenario_def,
                data_created=self._data_created)
            with open(
                    file=report_filename,
                    encoding="utf-8",
                    mode="w") as outfile:
                outfile.write(output)
        except jinja2.exceptions.TemplateSyntaxError as e:
            logging.error("Syntax error with preview.jinja")
            # for debugging/troubleshooting uncomment
            # raise e

    def generate_report(self, experiment_name: str) -> None:
        """Generates a preview of an experiment and writes that
        preview to disk.

        :param experiment_name: The name of the experiments
        :type experiment_name: str
        """
        folders = FolderManager(
            name=experiment_name,
            settings_file_override=self.settings_file_override)

        measurement_html = "<div><h1>Measurements</h1>"
        for engine in self.engines:
            # TODO: refactor this method to use the new settings file design
            # or pass in a folder manager from this object. For now not
            # supporting the settings file override
            measurement_html += engine.create_html_and_export(
                experiment_name=experiment_name, base_url=self.base_url,
                header=2)

        data_html = self._create_data_section(preview=False)

        report_folder = os.path.join(folders.base_folder, self.report_folder)

        if not os.path.exists(report_folder):
            os.makedirs(report_folder, exist_ok=True)

        report_filename = os.path.join(report_folder, self.report_filename)

        env = Environment(loader=FileSystemLoader('templates'))
        template = env.get_template("report.jinja")
        date_ran = get_timestamp_now(include_time=True)

        # if no tools then set to None so the template can report none
        tools:Union[list, None] = self._tools
        if len(self._tools) == 0:
            tools = []
        
        try:
            output = template.render(
                date_ran=date_ran,
                experiment_name=experiment_name,
                comments=self._comments,
                tools=tools,
                measurement_section=measurement_html,
                scenario_def=self._scenario_def,
                data_section=data_html)
            with open(
                    file=report_filename,
                    encoding="utf-8",
                    mode="w") as outfile:
                outfile.write(output)
        except jinja2.exceptions.TemplateSyntaxError as e:
            logging.error("Syntax error with report.jinja")
            # for debugging/troubleshooting uncomment
            # raise e

    def _create_html_link(
        self,
        description: str,
        file_w_path: str
    ) -> str:

        url_str = urljoin(base=self.base_url, url=file_w_path)
        return f"<a href={url_str}>{description}</a>"

    def _create_data_section(self, preview: bool = False) -> str:
        html = """<div>
        <h2>Data</h2>
        """
        if len(self._data_global) > 0:
            html += r"<h3>global</h3>"
            for entry in self._data_global:
                html += r"<li>"
                html += entry
                html += r"</li>"

        for component in self._data_sample:
            html += f"<h3>{component}<h3>\n<h4>Created</h4><ul>\n"
            # Created
            for entry in self._data_created[component]:
                html += r"<li>"
                html += entry
                html += r"</li>"
            # Samples
            if len(self._data_sample[component]) > 0:
                html += r"</ul><h4>Samples</h4><ul>"
                for entry in self._data_sample[component]:
                    html += r"<li>"
                    html += entry
                    html += r"</li>"
            # Metrics
            if len(self._data_metrics[component]) > 0:
                html += r"</ul><h4>Other</h4><ul>"
                for entry in self._data_metrics[component]:
                    html += r"<li>"
                    html += entry
                    html += r"</li>"
            html += r"</ul></div>"
        
            # Other
            if len(self._data_other[component]) > 0:
                html += r"</ul><h4>Other</h4><ul>"
                for entry in self._data_other[component]:
                    html += r"<li>"
                    html += entry
                    html += r"</li>"
            html += r"</ul></div>"
        
        return html

    def log_data_write(
        self,
        description: str,
        file_w_path: str,
        entry_type: Literal["sample", "created", "other", "metrics"],
        component: Union[str, None] = None
    ) -> None:
        logging.info(f"data write: {description} {file_w_path}")
        if component is not None:
            # ensure consistency across data logs
            if component not in self._data_sample:
                self._data_sample[component] = []
            if component not in self._data_created:
                self._data_created[component] = []
            if component not in self._data_other:
                self._data_other[component] = []
            if component not in self._data_metrics:
                self._data_metrics[component] = []                

        # now log
        html = f"{description} "
        html += self._create_html_link(
            description=file_w_path, file_w_path=file_w_path)

        if component is None:
            if html not in self._data_global:
                self._data_global.append(html)
        elif entry_type == "sample":
            if html not in self._data_sample[component]:
                self._data_sample[component].append(html)
        elif entry_type == "created":
            if html not in self._data_created[component]:
                self._data_created[component].append(html)
        elif entry_type == "metrics":
            if html not in self._data_metrics[component]:
                self._data_metrics[component].append(html)
        elif entry_type == "other":
            if html not in self._data_other[component]:
                self._data_other[component].append(html)

    def add_entry(
        self,
        entry_type: Literal["comment", "scenario_def", "measurement", "tool"],
        entry_text: str,
        component: Union[str, None] = None,
        include_timestamp: bool = False
    ) -> None:

        # if a scenario definition then no formatting or appending
        if entry_type == "scenario_def":
            self._scenario_def = entry_text
            return None

        # if its not a scenario definition then add the data
        entry = ""
        if include_timestamp:
            entry += get_timestamp_now()

        if component is not None:
            if self.use_formatting:
                entry = f"  <b>{component} : </b>"
            else:
                entry = f"  {component} : "

        entry += entry_text

        # avoid duplicate entries for reporting benefits. check here
        # just in case duplicate messages are sent from the objects such
        # as an Agent logging each loop the same "Save to file"  etc.
        if entry_type == "comment":
            if entry not in self._comments:
                self._comments.append(entry)
        elif entry_type == "measurement":
            if entry not in self._measurements:
                self._measurements.append(entry)
        elif entry_type == "tool":
            if entry not in self._tools:
                self._tools.append(entry)


"""metrics.py provides utility functions to simplify the process of 
retrieving metric results from measurement engine results.

This module includes functions to extract metric data from CSV files 
generated by measurement engines for individual runs or all runs in 
an experiment.
"""
import csv
import os
from typing import Union
from adgtk.tracking.structure import EXP_RESULTS_FOLDER

def get_single_run_metric_data(
    results_folder: str,
    engine_id: str,
    metric: str
) -> list:
    """
    Retrieve metric data for a single run from a CSV file.

    Args:
        results_folder (str): Path to the folder containing the results.
        engine_id (str): Identifier for the measurement engine.
        metric (str): Name of the metric to retrieve.

    Returns:
        list: A list of numeric values (int or float) representing the metric data. 
              Returns an empty list if the file does not exist or cannot be read.
    """
    meas_file = f"{engine_id}.{metric}.csv"
    file_w_path = os.path.join(results_folder, "metrics", meas_file)
    if not os.path.exists(file_w_path):
        return []
    data:list[Union[int, float]] = []
    try:
        with open(file_w_path, mode="r", encoding="utf-8") as infile:
            csv_reader = csv.reader(infile)
            for row in csv_reader:
                for entry in row:
                    try:
                        # Try converting to int first
                        data.append(int(entry))
                    except ValueError:
                        try:
                            # If int fails, try converting to float
                            data.append(float(entry))
                        except ValueError:
                            # Skip invalid entries
                            pass
    except (OSError, IOError):
        # Handle file read errors
        return []
    return data

def get_all_metric_data(experiment_name: str, engine_id: str, metric: str) -> dict:
    """
    Retrieve metric data for all runs in an experiment.

    Args:
        experiment_name (str): the name of the experiment.
        engine_id (str): Identifier for the measurement engine.
        metric (str): Name of the metric to retrieve.

    Returns:
        dict: A dictionary where keys are run folder names and values 
              are lists of numeric values (int or float) representing 
              the metric data for each run. Returns an empty dictionary 
              if no valid directories are found.
    """
    results = {}
    experiment_root = os.path.join(EXP_RESULTS_FOLDER, experiment_name)
    run_folders = [folder for folder in os.listdir(experiment_root) 
                   if os.path.isdir(os.path.join(experiment_root, folder))]
    for folder in run_folders:
        folder_w_path = os.path.join(experiment_root, folder)
        results[folder] = get_single_run_metric_data(
            results_folder=folder_w_path,
            engine_id=engine_id,
            metric=metric)
    return results


"""a Journal provides reports and other summaries of an experiment.

Versions:
v 0.1
- mvp

Defects:
- linter is throwing a  E0401 which is a known bug. Does not impact
  execution. Its only an editor issue.
"""


# ---------------------------------------------------------------------
# Imports
# ---------------------------------------------------------------------
__all__ = ["experiment"]

from .experiment import ExperimentJournal
from .project import ProjectJournal
from .base import SupportsReportingOperations



"""Provides a central measurement factory.

Goals
=====
a dynamic factory that is focused on measuring of data.

Design
======
1. manages both classes as well as functions.
2. if its a class, supports both the class and an instance of
3. if its a class, the init must not require arguments. If you need
   to modify the init, use the full component factory
4. The goal is to centralize the measurements, both  those built-in and
   user defined.

Testing
=======
Pending

Notes
=====
1. 
"""
import os
import sys
# ----------------------------------------------------------------------
# Start of path verification
# ----------------------------------------------------------------------
path = os.getcwd()
bootstrap_file = os.path.join(path, "bootstrap.py")
if not os.path.exists(bootstrap_file):
    print("ERROR: Unable to locate the bootstrap.py. Please check your path.")
    sys.exit(1)
# ----------------------------------------------------------------------
# End of path verification
# ----------------------------------------------------------------------
import inspect
from typing import (
    Any,
    Callable,
    Iterable,
    Literal,
    Optional,
    Protocol,
    Sequence,
    TypedDict,
    Union,
    get_args,
    runtime_checkable
)
import uuid
import numpy as np
from adgtk.utils import create_logger
# ----------------------------------------------------------------------
# Protocols, Interfaces, & Structures
# ----------------------------------------------------------------------

# ------------- measurements -------------
@runtime_checkable
class ClassBasedMeasurement(Protocol):
    """For generic class based measurements"""

    def __call__(self, a:Any) -> Union[int, float]: ...


@runtime_checkable
class ClassBasedComparison(Protocol):
    """For generic class based comparisons"""

    def __call__(self, a:Any, b:Any) -> Union[int, float]: ...

direct_measurement = Callable[[Any], Union[int, float]]
direct_comparison = Callable[[Any, Any], Union[int, float]]
distribution_measurement = Callable[[Any], Union[list, np.ndarray]]
distribution_comparison = Callable[[Iterable, Iterable], float]

supports_factory = Union[
        ClassBasedComparison,
        ClassBasedMeasurement,
        direct_measurement,
        direct_comparison,
        distribution_measurement,
        distribution_comparison
        ]

measurement_type = Literal[
        "class_based_measure",
        "class_based_compare",
        "class_other",        
        "direct_measure",
        "direct_comparison",
        "direct_other",
        "distribution_measure",
        "distribution_comparison"]

# ------------- entry -------------

class MeasFactoryEntry(TypedDict):
    factory_id:str
    meas_type: measurement_type
    tags: list[str]
    item: supports_factory
    description: str


# ----------------------------------------------------------------------
# Globals
# ----------------------------------------------------------------------

_inventory: dict[str, MeasFactoryEntry] = {}


# Set up module-specific logger
_logger = create_logger(
    "adgtk.measurement.factory.log",
    logger_name=__name__,
    subdir="framework"
)


# ----------------------------------------------------------------------
# Helpers, can be public but designed for internal                                                            
# ----------------------------------------------------------------------
def classify_measurement(item: Callable) -> measurement_type:
    if inspect.isclass(item):
        if issubclass(item, ClassBasedMeasurement):
            return "class_based_measure"
        elif issubclass(item, ClassBasedComparison):
            return "class_based_compare"
        else:
            return "class_other"                    
    elif inspect.isfunction(item):
        if get_args(item) == get_args(direct_measurement):
            return "direct_measure"
        if get_args(item) == get_args(direct_comparison):
            return "direct_comparison"        
        if get_args(item) == get_args(distribution_measurement):
            return "distribution_measure"
        if get_args(item) == get_args(distribution_comparison):
            return "distribution_comparison"
        return "direct_other"
    raise ValueError("Not callable")

# ----------------------------------------------------------------------
# Public, designed for use outside module
# ----------------------------------------------------------------------
def get_measurement_factory_entry(factory_id:str) -> MeasFactoryEntry:
    """Gets the entry for a factory id

    Args:
        factory_id (str): the id to get

    Raises:
        IndexError: Unable to locate the factory id requested

    Returns:
        MeasFactoryEntry: The entry in the factory
    """
    if factory_id not in _inventory.keys():
        _logger.error("unable to locate factory_id %s", factory_id)
        raise IndexError("unable to locate factory_id %s", factory_id)    
    return _inventory[factory_id]


def register_to_measurement_factory(
        description: Optional[str] = None,
        tags: Optional[list[str]] = None,
        factory_id: Optional[str] = None
    ):
    """Decorator that registers a measurement to the factory

    Args:
        description (Optional[str]): Overrides the docstring
        tags (Optional[list[str]]): For filtering by user tag
        factory_id (Optional[str]): Requesting a specific ID

    Raises:
        ValueError: No docstring or description set
    """
    def decorator(cls):
        nonlocal description
        if description is None:
            description = inspect.getdoc(cls)
        if description is None:
            _logger.error(
                "Attempted to register via decorator without a "
                "description or docstring")
            raise ValueError(
                "register_to_factory requires a docstring or description set")
        _ = manual_measurement_factory_register(
            item=cls,
            description=description,
            tags=tags,
            factory_id=factory_id
        )
        return cls

    return decorator
    

def manual_measurement_factory_register(
    item: supports_factory,
    description:str,
    tags: Optional[list[str]] = None,
    factory_id: Optional[str] = None,
) -> str:
    """registers a measurement to the measurement factory

    Args:
        item (supports_factory): The item to register
        description (str): The associated description
        tags (Optional[list[str]], optional): Tags for further filtering
            Defaults to None.
        factory_id (Optional[str], optional): The requested ID
            Defaults to None.

    Raises:
        KeyError: The requested factory ID already exists

    Returns:
        str: the registered factory id
    """
    global _inventory
    if factory_id is None:
        if inspect.isfunction(item) or inspect.isclass(item):
            factory_id = item.__name__
        else:
            factory_id = str(uuid.uuid4())
    if factory_id in _inventory.keys():
        _logger.error(
            "Factory Id %s already exists. unable to register", factory_id)
        raise KeyError("Factory Id %s already exists", factory_id)
    if tags is None:
        tags = []
    
    entry = MeasFactoryEntry(
        item=item,
        tags=tags,
        description=description,
        meas_type=classify_measurement(item),
        factory_id=factory_id
    )    
    _inventory[factory_id] = entry
    _logger.info("Registered %s to factory", factory_id)
    return factory_id


def get_measurements_by_type(
    filter_by_type:Optional[measurement_type] = None
) -> list[MeasFactoryEntry]:
    """Returns the measurements in the factory based on type

    Args:
        filter_by_type (Optional[measurement_type], optional): Filter by
            type. Defaults to None.

    Returns:
        list[MeasFactoryEntry]: All factory entries that fit the search
    """
    if filter_by_type is None:
        return list(_inventory.values())
    entries = []
    for _, item in _inventory.items():
        if item["meas_type"] == filter_by_type:
            entries.append(item)
    return entries

def get_measurements_by_tag(
    tags:Optional[Union[str, list[measurement_type]]]
) -> list[MeasFactoryEntry]:
    """Returns the measuments in the factory based on tag

    Args:
        tags (Optional[Union[str, list[measurement_type]]]): _description_

    Returns:
        list[MeasFactoryEntry]: _description_
    """
    
    if tags is None:
        return list(_inventory.values())
    if isinstance(tags, str):
        search_tags = [tags]
    elif isinstance(tags, Sequence):
        search_tags = tags    
    entries = []
    for _, item in _inventory.items():
        # check for overlap between tags
        if bool(set(search_tags) & set(item["tags"])):
            entries.append(item)
    return entries


def create_measurement(factory_id:str, **kwargs) -> supports_factory:
    """Creates the measurement for calling purposes. If already callable
    it returns the registered measurement.

    Args:
        factory_id (str): The factory ID to create
        **kwargs: Additional arguments to pass to the class constructor
                  if the item is a class.

    Raises:
        IndexError: If the factory ID is not found.

    Returns:
        supports_factory: The created or registered measurement.
    """
    if factory_id not in _inventory.keys():
        raise IndexError("factory id %s not found", factory_id)
    
    entry = _inventory[factory_id]

    # direct
    if entry["meas_type"].startswith("direct"):
        return entry["item"]
    
    # class based
    item = entry["item"]
    if inspect.isclass(item):
        return item(**kwargs)    
    return item


def get_measurement_factory_labels() -> list[str]:
    return list(_inventory.keys())



"""Impairment Engine is responsible for impairing the data. It follows
a similar pattern as the measurement engine."""

from typing import Protocol, Union, Literal, runtime_checkable
from adgtk.common import DuplicateFactoryRegistration, InvalidConfigException
from adgtk.components import PresentableRecord, PresentableGroup, State
# ----------------------------------------------------------------------
#
# ----------------------------------------------------------------------
# py -m pytest -s test/impairment/test_engine.py


@runtime_checkable
class Impairment(Protocol):
    """Impairment is a protocol for impairing data"""

    impairs: set[
        Literal["PresentableGroup", "PresentableRecord", "dict", "str", "State"]
    ]
    label: str
    skipped_counter: int
    executed_counter: int

    def impair(
        self,
        data: Union[PresentableGroup, PresentableRecord, dict, str, State],
        key: Union[str, None] = None,
        idx: Union[int, None] = None
    ) -> dict:
        """Impair the data

        :param data: The data to impair
        :type data: Union[PresentableGroup, PresentableRecord, dict, str, State]
        :param key: If a specific key is requested to impair, defaults to None
        :type key: Union[str, None], optional
        :param idx: If a group then the record to impair, defaults to None
        :type idx: Union[int, None], optional
        :return: the data impaired
        :rtype: dict
        """


class ImpairmentEngine:
    """Impairment Engine is responsible for impairing the data. It follows
    a similar pattern as the measurement engine."""

    def __init__(self):
        """Initializes a new instance of the Impairment Engine."""
        self._impairments: dict[str, Impairment] = {}

    def add_impairment(self, impairment: Impairment) -> None:
        """Adds an impairment to the engine.

        :param impairment: The impairment to add
        :type impairment: Impairment
        """
        if not isinstance(impairment, Impairment):
            msg = "Impairment must be an instance of Impairment"
            raise ValueError(msg)

        if impairment.label not in self._impairments:
            self._impairments[impairment.label] = impairment
        else:
            msg = f"Impairment {impairment.label} exists"
            raise DuplicateFactoryRegistration(msg)

    def __len__(self) -> int:
        """Returns the number of impairments in the engine

        :return: The number of impairments
        :rtype: int
        """
        return len(self._impairments)

    def _can_impair(
        self,
        impairs: set[
            Literal[
                "State",
                "PresentableGroup",
                "PresentableRecord",
                "dict",
                "str"]],
        data: Union[PresentableGroup, PresentableRecord, dict, str, State]
    ) -> bool:
        """Internal function to check if the data can be impaired

        :param impairs: The types of data that can be impaired
        :type impairs: list[Literal["PresentableGroup", "PresentableRecord", "dict", "str", "State"]]
        :param data: The data to check
        :type data: Union[PresentableGroup, PresentableRecord, dict, str, State]
        :return: True if the data can be impaired
        :rtype: bool
        """
        for imp in impairs:
            if self._check_type(imp, data):
                return True
        return False

    def _check_type(
        self,
        impairs:
            Literal[
                "State",
                "PresentableGroup",
                "PresentableRecord",
                "dict",
                "str"],
        data: Union[PresentableGroup, PresentableRecord, dict, str, State]
    ) -> bool:
        """Checks against a specific type. Used by _can_impair

        :param impairs: The type of data that can be impaired
        :type impairs: Literal[ "PresentableGroup", "PresentableRecord", "dict", "str", "State"]
        :param data: The data to check
        :type data: Union[PresentableGroup, PresentableRecord, dict, str, State]
        :return: True if the data can be impaired
        :rtype: bool
        """
        match impairs:
            case "PresentableGroup":
                return isinstance(data, PresentableGroup)
            case "PresentableRecord":
                return isinstance(data, PresentableRecord)
            case "dict":
                return isinstance(data, dict)
            case "str":
                return isinstance(data, str)
            case _:
                return False

    def impair(
        self,
        impairment_label: str,
        data: Union[PresentableGroup, PresentableRecord, dict, str, State],
        key: Union[str, None] = None,
        idx: Union[int, None] = None
    ) -> dict:
        """Impair the data. Assumes working off a copy.

        :param impairment_label: The label of the impairment to use
        :type impairment_label: str
        :param data: The data to impair
        :type data: Union[PresentableGroup, PresentableRecord, dict, str]
        :param key: If a specific key is requested to impair, defaults to None
        :type key: Union[str, None], optional
        :param idx: If a group then the record to impair, defaults to None
        :type idx: Union[int, None], optional
        :return: a copy of the data impaired
        :rtype: dict
        """
        if impairment_label in self._impairments:
            impair_type = self._impairments[impairment_label].impairs
            if self._can_impair(impair_type, data):
                return self._impairments[impairment_label].impair(
                    data=data, key=key, idx=idx)
            msg = f"Impairment {impairment_label} cannot impair {type(data)}"
            raise InvalidConfigException(msg)
        else:
            msg = f"Impairment {impairment_label} does not exist"
            raise ValueError(msg)

    def get_impairment_labels(
        self,
        impairs: Literal[
            "PresentableGroup",
            "PresentableRecord",
            "dict",
            "str",
            "all"]
    ) -> list[str]:
        """Returns the labels of the registered impairments. Usage
        includes for generating reports and needing to iterate the
        impairment engine.
        :param impairs: The type of impairments to return
        :type impairs: Literal["PresentableGroup", "PresentableRecord", "dict", "str", "all"]

        :return: A list of the impairment labels
        :rtype: list
        """
        if impairs == "all":
            return list(self._impairments.keys())
        else:
            return_list = []
            for label, imp in self._impairments.items():
                if impairs in imp.impairs:
                    return_list.append(label)
            return return_list

        return []

    def get_impairment(self, label: str) -> Impairment:
        """Returns the impairment associated with the label

        :param label: The label of the impairment
        :type label: str
        :return: The impairment
        :rtype: Impairment
        """
        if label in self._impairments:
            return self._impairments[label]
        else:
            msg = "Impairment %s does not exist", label
            raise ValueError(msg)

    def get_skip_count(self, label: str) -> int:
        """Gets the number of times the impairment was skipped

        :param label: the label of the impairment
        :type label: str
        :return: the counter value
        :rtype: int
        """
        if label in self._impairments:
            return self._impairments[label].skipped_counter

        raise KeyError(f"Impairment {label} does not exist")

    def get_execution_count(self, label: str) -> int:
        """Gets the number of times the impairment was executed

        :param label: the label of the impairment
        :type label: str
        :return: the counter value
        :rtype: int
        """
        if label in self._impairments:
            return self._impairments[label].executed_counter

        raise KeyError(f"Impairment {label} does not exist")

    def reset_counters(self, label: Union[str, None] = None) -> None:
        """Resets the counters for the impairments

        :param label: filter the impairment counter to reset, defaults to None
        :type label: Union[str, None], optional
        """

        if label is not None:
            if label in self._impairments:
                self._impairments[label].skipped_counter = 0
                self._impairments[label].executed_counter = 0
        else:
            for key in self._impairments.keys():
                self._impairments[key].skipped_counter = 0
                self._impairments[key].executed_counter = 0



# mypy: ignore-errors
"""Policy protoocols provides the foundational components for creating
your own policy.
"""


from typing import List, Protocol, runtime_checkable, Union
from adgtk.common import FactoryBlueprint
from adgtk.components import Action, State, StateType
from adgtk.tracking import PerformanceTracker


@runtime_checkable
class Policy(Protocol):
    """The Policy base"""
    blueprint: FactoryBlueprint
    training: bool
    supports_env_type: List[StateType]
    
    # for policy updates, tracking
    experiment_name: Union[str, None] = None
    performance_tracker: Union[PerformanceTracker, None] = None


    def reset(self) -> None:
        """Resets internal state during training."""

    def invoke(self, state: State) -> Action:
        """Invoke the policy to include tracking for training. The
        policy can chose to explore or exploit in response to the ask.

        Args:
            state (State): The state

        Returns:
            Action: The action to take
        """

    def sample(self, state: State) -> Action:
        """Invokes the policy but does not update for training. It only
        seeks to exploit.

        Args:
            state (State): The state

        Returns:
            Action: The action to take
        """
        pass

    def update(self, reward: float) -> None:
        """Updates a policy using the reward from the environment for
        the last action.

        :param reward: The reward from the last action
        :type reward: float
        """

    def refresh(self) -> None:
        """Refreshes the policy by creating training data based on the
        last epoch. This will be used when there is a model to train.
        If there is nothing to update "refresh" then this is a no-op.
        """


@runtime_checkable
class UsesPrompts(Protocol):
    """For logging of the last prompt we need to confirm its implemented"""

    last_prompt: str
